import re

from metaerg import context
from metaerg.datatypes import fasta
from metaerg.datatypes import sqlite


def _run_programs(genome_name, contig_dict, db_connection, result_files):
    fasta_file = context.spawn_file('masked', genome_name)
    fasta.write_contigs_to_fasta(contig_dict, fasta_file, db_connection, genome_name,
                                      mask_targets=fasta.ALL_MASK_TARGETS)
    context.run_external(f'aragorn -l -t -gc{context.TRANSLATION_TABLE} {fasta_file} -w -o {result_files[0]}')

def _read_results(genome_name, contig_dict, db_connection, result_files) -> int:
    count = 0
    ln = 0
    current_contig = None
    coord_regexp = re.compile(r'(c*)\[(\d+),(\d+)]')
    with open(result_files[0]) as aragorn_handle:
        for line in aragorn_handle:
            words = line.strip().split()
            ln += 1
            match words:
                case ['>end']:
                    break
                case [contig_name] if contig_name.startswith('>'):
                    if space_index := contig_name.find(' ') > 0:
                        contig_name = contig_name[:space_index]
                    current_contig = contig_dict[contig_name[1:]]
                case [_, trna, coordinates, _, codon]:
                    if coord_match := coord_regexp.fullmatch(coordinates):
                        if current_contig:
                            strand = -1 if 'c' == coord_match.group(1) else 1
                            start = max(0, int(coord_match.group(2)) - 1)
                            end = min(len(current_contig['seq']), int(coord_match.group(3)))
                            seq = current_contig['seq'][start:end]
                            if strand < 0:
                                seq = fasta.reverse_complement(seq)
                            feature = sqlite.Feature(genome = genome_name,
                                       contig = current_contig['id'],
                                       start = start,
                                       end = end,
                                       strand = strand,
                                       type = 'tRNA',
                                       inference = 'aragorn',
                                       nt_seq = seq,
                                       descr = f'{trna}-{codon}')
                            sqlite.add_new_feature_to_db(db_connection, feature)
                            count += 1
                        else:
                            context.log(f'Warning: Unexpected tRNA coordinates in "{line}" '
                                    f'at line {ln - 1} of {result_files[0]}')
    return count


@context.register_annotator
def run_and_read_aragorn():
    return ({'pipeline_position': 11,
             'annotator_key': 'aragorn',
             'purpose': 'tRNA prediction with aragorn',
             'programs': ('aragorn',),
             'result_files': ("aragorn",),
             'run': _run_programs,
             'read': _read_results})




import re

from metaerg import context
from metaerg.datatypes import fasta
from metaerg.datatypes import sqlite


def _run_programs(genome_name, contig_dict, db_connection, result_files):
    fasta_file = context.spawn_file('masked', genome_name)
    # no masking here becasuse we want to arbitrate with repeatmasker results
    if context.TRANSLATION_TABLE < 0:
        translate = '-p meta'
    else:
        translate = f'-g {context.TRANSLATION_TABLE}'
    context.run_external(f'prodigal {translate} -m -f gff -q -i {fasta_file} -a {result_files[0]} -d {result_files[1]}')

def _read_results(genome_name, contig_dict, db_connection, result_files) -> int:
    ORF_ID_PATTERN = re.compile(r'_(\d+?)$')
    nucl_seq_hash = {}
    with fasta.FastaParser(result_files[1], cleanup_seq=False) as fasta_reader:
        for seq_rec in fasta_reader:
            nucl_seq_hash[seq_rec['id']] = seq_rec['seq']
    count = 0
    dropped_repeat_count = 0
    repeat_count_before_arbitration = sum(1 for f in sqlite.read_all_features(db_connection, type='repeat'))
    with fasta.FastaParser(result_files[0], cleanup_seq=False) as fasta_reader:
        rejected_cds_count = 0
        for seq_rec in fasta_reader:
            x_count = seq_rec['seq'].count('X')
            if x_count > 0.2 * len(seq_rec['seq'] or x_count >= 10):
                rejected_cds_count += 1
                continue
            words = seq_rec['descr'].split('#')
            try:
                m = ORF_ID_PATTERN.search(seq_rec['id'])
                contig_id = seq_rec['id'][0:m.start()]
            except KeyError:
                context.log(f'({genome_name}) Warning: Failed to find contig with "{seq_rec["id"]}"')
                continue
            start = int(words[1].strip()) - 1
            end = int(words[2].strip())

            # reconciliation with repeats
            overlapping_features = [f for f in sqlite.read_all_features(db_connection, contig=contig_id,
                                    location=(start, end), type='rRNA tRNA tmRNA ncRNA crispr_repeat repeat retrotransposon'.split())]

            if len(overlapping_features):
                overlap = 0
                for f in overlapping_features:
                    overlap += min(f.end, end) - max(start, f.start)
                non_repeat_features = [f for f in overlapping_features if f.type in 'rRNA tRNA tmRNA ncRNA crispr_repeat retrotransposon'.split()]
                if len(non_repeat_features):
                    rejected_cds_count += 1
                    continue
                elif overlap < 0.33 * (end-start):
                    for overlapping_feature in overlapping_features:
                        sqlite.drop_feature(db_connection, overlapping_feature)
                        dropped_repeat_count += 1
                else:
                    rejected_cds_count += 1
                    continue

            strand = int(words[3].strip())
            if seq_rec['seq'].endswith('*'):
                seq_rec['seq'] = seq_rec['seq'][:-1]
            feature = sqlite.Feature(genome = genome_name,
                       contig = contig_id,
                       start = start,
                       end = end,
                       strand = strand,
                       type = 'CDS',
                       inference = 'prodigal',
                       aa_seq = seq_rec['seq'],
                       nt_seq = nucl_seq_hash[seq_rec['id']])
            if 'partial=01' in seq_rec['descr'] or 'partial=01' in seq_rec['descr'] or 'partial=11' in seq_rec['descr']:
                feature.notes = 'partial protein'
            sqlite.add_new_feature_to_db(db_connection, feature)
            count += 1

        context.log(f'({genome_name}) Dropped {dropped_repeat_count}/{repeat_count_before_arbitration} repeats and'
                    f' rejected {rejected_cds_count} CDS during arbitration of prodigal results.')
        return count


@context.register_annotator
def run_and_read_prodigal():
    return ({'pipeline_position': 61,
             'annotator_key': 'prodigal',
             'purpose': 'coding sequence prediction with prodigal',
             'programs': ('prodigal',),
             'result_files': ('prodigal','prodigal-nucl'),
             'run': _run_programs,
             'read': _read_results})



import sqlite3 as sql

from metaerg.datatypes.blast import BlastResult, BlastHit, DBentry

FEATURE_FIELDS = tuple('id genome contig start end strand type inference subsystems descr taxon notes ' \
                    'aa_seq nt_seq antismash signal_peptide tmh tmh_topology blast cdd hmm'.split())
RNA_TARGETS = set("rRNA tRNA tmRNA ncRNA retrotransposon".split())
SQLITE_CREATE_TABLE_SYNTAX = '''CREATE TABLE features(
    id TEXT,
    genome TEXT,
    contig TEXT,
    start INT,
    end INT,
    strand INT,
    type TEXT,
    inference TEXT,
    subsystems TEXT,
    descr TEXT,
    taxon TEXT,
    notes TEXT,
    aa_seq TEXT,
    nt_seq TEXT,
    antismash TEXT,
    signal_peptide TEXT,
    tmh INT,
    tmh_topology TEXT,
    blast TEXT,
    cdd TEXT,
    hmm TEXT
)'''
SQLITE_UPDATE_SYNTAX = '''UPDATE features SET
    id = ?,
    genome = ?,
    contig = ?,
    start = ?,
    end = ?,
    strand = ?,
    type = ?,
    inference = ?,
    subsystems = ?,
    descr = ?,
    taxon = ?,
    notes = ?,
    aa_seq = ?,
    nt_seq = ?,
    antismash = ?,
    signal_peptide = ?,
    tmh = ?,
    tmh_topology = ?,
    blast = ?,
    cdd = ?,
    hmm = ? 
WHERE rowid = ?'''

class Feature:
    def __init__(self,
                 rowid: int = 0,
                 id: str = '',
                 genome: str = '',
                 contig: str = '',
                 start: int = 0,
                 end: int = 0,
                 strand: int = 0,
                 type: str = '',
                 inference: str = '',
                 subsystems: str = '',
                 descr: str = '',
                 taxon: str = '',
                 notes: str = '',
                 aa_seq: str = '',
                 nt_seq: str = '',
                 antismash: str = '',
                 signal_peptide: str = '',
                 tmh: int = 0,
                 tmh_topology: str = '',
                 blast: str = '',
                 cdd: str = '',
                 hmm: str = ''):
        self.rowid = rowid
        self.id = id
        self.genome = genome
        self.contig = contig
        self.start = start
        self.end = end
        self.strand = strand
        self.type = type
        self.inference = inference
        self.subsystems = subsystems
        self.descr = descr
        self.taxon = taxon
        self.notes = notes
        self.aa_seq = aa_seq
        self.nt_seq = nt_seq
        self.antismash = antismash
        self.signal_peptide = signal_peptide
        self.tmh = tmh
        self.tmh_topology = tmh_topology
        self.blast = eval(blast) if blast else None
        self.cdd = eval(cdd) if cdd else None
        self.hmm = eval(hmm) if hmm else None

    def __iter__(self):
        for k, v in self.__dict__.items():
            if k == 'rowid':
                continue
            yield k, v

    def __repr__(self):
        return '{}({})'.format(type(self).__name__, ', '.join(f'{k}={v!r}' for k, v in self))

    def __str__(self):
        return '{}:\n{}'.format(type(self).__name__, '\n'.join(f'  {k:14}: {v}' for k, v in self))

    def __len__(self):
        return len(self.__dict__.items())

    def length_nt(self):
        return self.end - self.start

    def length_aa(self):
        return (self.end - self.start) // 3

def feature_factory(cursor, row) -> Feature:
    fields = [column[0] for column in cursor.description]
    if len(fields) < 10:
        return fields[0]
    return Feature(**{key: value for key, value in zip(fields, row)})

def connect_to_db(sql_db_file):
    connection = sql.connect(sql_db_file)
    connection.row_factory = feature_factory
    return connection

def create_db(sql_db_file):
    sql_db_file.unlink(missing_ok=True)
    connection = sql.connect(sql_db_file)
    cursor = connection.cursor()
    cursor.execute(SQLITE_CREATE_TABLE_SYNTAX)

def add_new_feature_to_db(sql_connection, feature: Feature):
    feature_as_tuple = tuple(str(v) for k, v in feature)
    cursor = sql_connection.cursor()
    cursor.execute('INSERT INTO features VALUES(?,?,?,?, ?,?,?,?, ?,?,?,?, ?,?,?,?, ?,?,?,?, ?)', feature_as_tuple)
    sql_connection.commit()

def update_feature_in_db(sql_connection, feature: Feature):
    cursor = sql_connection.cursor()
    feature_as_list = list(str(v) for k, v in feature)
    feature_as_list.append(feature.rowid)
    cursor.execute(SQLITE_UPDATE_SYNTAX, tuple(feature_as_list))
    sql_connection.commit()

#def update_feature_id_in_db(sql_connection, feature: Feature):
#    cursor = sql_connection.cursor()
#    cursor.execute('UPDATE features SET id = ? WHERE genome = ? AND contig = ? AND start = ? AND end = ? AND '
#                   'strand = ? AND type = ? AND inference = ?',
#                   (feature.id, feature.genome, feature.contig, feature.start, feature.end, feature.strand,
#                    feature.type, feature.inference))
#    sql_connection.commit()

def drop_feature(sql_connection, feature: Feature):
    cursor = sql_connection.cursor()
    cursor.execute('DELETE FROM features WHERE rowid = ?',
                   (feature.rowid,))
    sql_connection.commit()

def count_features(sql_connection):
    sql_connection.row_factory = None
    cursor = sql_connection.cursor()
    total = sum(1 for f in cursor.execute('SELECT start FROM features'))
    sql_connection.row_factory = feature_factory
    return total

def read_feature_by_id(sql_connection, feature_id) -> Feature:
    cursor = sql_connection.cursor()
    result = cursor.execute('SELECT rowid, * FROM features WHERE id = ?', (feature_id,))
    return result.fetchone()

def read_all_features(sql_connection, contig='', type=None, location=None, additional_sql=None):
    cursor = sql_connection.cursor()

    where_str = []
    fields = []
    if type:
        if isinstance(type, list) or isinstance(type, tuple)  or isinstance(type, set):
            where_str.append('type IN ({})'.format(','.join('?' for t in type)))
            fields.extend(type)
        else:
            where_str.append('type = ?')
            fields.append(type)
    if contig:
        where_str.append('contig = ?')
        fields.append(contig)
    if location:
        where_str.append('start < ? AND ? < end')
        fields.append(location[1])
        fields.append(location[0])
    if additional_sql:
        where_str.append(additional_sql)

    where_str = ' AND '.join(where_str)
    if where_str:
        #print('SELECT * FROM features WHERE ' + where_str + ' ORDER BY contig, start')
        #print(fields)
        return cursor.execute('SELECT rowid, * FROM features WHERE ' + where_str + ' ORDER BY contig, start', tuple(fields))
    else:
        return cursor.execute('SELECT rowid, * FROM features ORDER BY contig, start')
    #for columns in result.fetchall():
    #    yield Feature(*columns)



import re
import gzip
import textwrap

from metaerg.datatypes import sqlite

GBK_LINEWIDTH = 80
GBK_INDENT = 21
GBK_HEADER = '''LOCUS       {:<15} {:>12} bp    DNA              UNK 01-JAN-1980
DEFINITION  {}.
ACCESSION   {}
VERSION     {}
KEYWORDS    .
SOURCE      .
  ORGANISM  .
            .
FEATURES             Location/Qualifiers
'''


def gbk_write_feature(writer, feature: sqlite.Feature):
    indent = ' '*GBK_INDENT
    if feature.strand >= 0:
        location = f'{int(feature.start) + 1}..{int(feature.end)}'
    else:
        location = f'complement({int(feature.start) + 1}..{int(feature.end)})'

    writer.write('     {:<16}{}\n'.format(feature.type, textwrap.fill(location, width = GBK_LINEWIDTH,
                                                                      initial_indent = '',
                                                                      subsequent_indent = indent)))
    gbk_keys = {'locus_tag':  feature.id,
                'inference':  feature.inference,
                'product':    feature.descr,
                'taxonomy':   feature.taxon,
                'subsystems': feature.subsystems,
                'notes':      feature.notes,
                'antismash':  feature.antismash,
                'signal_peptide': feature.signal_peptide,
                'tmh':         feature.tmh,
                'tmh-topology':feature.tmh_topology,
                'translation': feature.aa_seq if feature.type == 'CDS' else ''}
    for k, v in gbk_keys.items():
        if v:
            writer.write(textwrap.fill(f'/{k}="{v}"', width = GBK_LINEWIDTH,
                                       initial_indent = indent, subsequent_indent = indent))
            writer.write('\n')


def gbk_write_genome(writer, contig_dict: dict, db_connection):
    for contig_id, contig in contig_dict.items():
        writer.write(GBK_HEADER.format(contig_id, len(contig['seq']), contig_id, contig_id, contig_id))
        for feature in sqlite.read_all_features(db_connection, contig=contig_id):
            gbk_write_feature(writer, feature)
        writer.write('ORIGIN\n')
        lw = ((GBK_LINEWIDTH - 20) // 10) * 10
        for i in range(0, len(contig['seq']), lw):
            line_seq = contig['seq'][i:i + lw].lower()
            writer.write(f'{i+1:>9}')
            for j in range(lw // 10):
                writer.write(' ')
                writer.write(line_seq[j*10:j*10+10])
            writer.write('\n')
        writer.write('//\n')


class GbkFeatureParser:
    def __init__(self, path):
        self.path = path
        self.handle = None
        self.feature_re = re.compile(r'\s{,6}(\w+)\s+(complement\()*(\d+)\.\.(\d+)\)*')
        self.qualifier_re = re.compile(r'/(\w+?)=(.+)')

    def __enter__(self):
        if str(self.path).endswith('.gz'):
            self.handle = gzip.open(self.path, 'rt')
        else:
            self.handle = open(self.path)
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.handle.close()

    def __iter__(self):
        current_contig_name = ''
        current_feature = {}
        current_k = ''
        current_v = ''
        parsing_features = False
        while line := self.handle.readline():
            if line.startswith('ACCESSION'):
                current_contig_name = line.split()[1]
                continue
            elif line.startswith('FEATURES'):
                parsing_features = True
                continue
            elif line.startswith('ORIGIN'):
                if current_feature:
                    if current_v:
                        current_feature[current_k] = current_v.strip('"')
                    yield current_feature
                parsing_features = False
                continue
            elif not parsing_features:
                continue
            line = line.strip('\n')
            if m := self.feature_re.match(line):
                if current_feature:
                    yield current_feature
                current_feature = {'contig': current_contig_name,
                                   'start': int(m.group(3))-1,
                                   'end': int(m.group(4)),
                                   'strand': -1 if m.group(2) else 1,
                                   'type': m.group(1)
                                   }
            elif current_feature:
                line =line.strip()
                if m := self.qualifier_re.fullmatch(line):
                    if current_v:
                        current_feature[current_k] = current_v.strip('"')
                    current_k = m.group(1)
                    current_v = m.group(2)
                else:
                    current_v += f' {line}'


def pad_seq(sequence):
    """ Pad sequence to multiple of 3 with N """
    remainder = len(sequence) % 3
    return sequence if remainder == 0 else sequence + 'N' * (3 - remainder)


import ast
import re
from pathlib import Path

from metaerg import context
from metaerg.datatypes import fasta

def write_html(dir):
    dir.mkdir(exist_ok=True, parents=True)
    file = Path(dir, 'index.html')
    with open(file, 'w') as handle:
        handle.write(make_html())


def make_html() -> str:
    """Injects the content into the html base, returns the html."""
    genome_list = []
    with open(context.GENOME_NAME_MAPPING_FILE) as handle:
        for line in handle:
            genome_list.append(line.split())

    checkm_results = {}
    checkm_result_file = Path(context.CHECKM_DIR, 'storage', 'bin_stats_ext.tsv')
    if checkm_result_file.exists():
        with open(checkm_result_file) as handle:
            for line in handle:
                words = line.split('\t')
                checkm_results[words[0]] = ast.literal_eval(words[1])

    gtdbtk_results = {}
    for file in (Path(context.GTDBTK_DIR, 'gtdbtk.ar53.summary.tsv'),
                 Path(context.GTDBTK_DIR, 'gtdbtk.bac120.summary.tsv')):
        if file.exists():
            with open(file) as handle:
                for line in handle:
                    if line.startswith("user_genome\t"):
                        continue
                    words = line.split("\t")
                    words[1] = re.sub('[a-z]__', ' ', words[1])
                    gtdbtk_results[words[0]] = words[1]

    html = _make_html_template()
    left_aligned = 'file name classification'
    headers = ''
    for column in 'file name (Mb) N50 code completeness contamination classification'.split():
        if column in left_aligned:
            headers += f'<th id=al>{column}</th>\n'
        else:
            headers += f'<th>{column}</th>\n'
    html = html.replace('HEADERS', headers)

    colors = 'id=cr id=cr id=co id=cb id=cg'.split()
    genomes_html = ''
    for new_name, old_name, contig_file in genome_list:
        with fasta.FastaParser(contig_file) as fasta_reader:
            contigs = [c for c in fasta_reader]
        contigs.sort(key=lambda c: len(c['seq']), reverse=True)
        genome_size = sum(len(c['seq']) for c in contigs) / 1e6
        N50 = 0
        for c in contigs:
            N50 += len(c)
            if N50 >= genome_size / 2:
                N50 = len(c)
                break
        completeness = 0
        contamination = 0
        completeness_as_str = ''
        contamination_as_str = ''
        code = ''
        gtdbtk_classification = ''
        try:
            checkm_result = checkm_results[old_name]
            if not checkm_result:
                checkm_result = checkm_results[new_name]
            if checkm_result:
                completeness = float(checkm_result["Completeness"])
                contamination = float(checkm_result["Contamination"])
                completeness_as_str = f'{completeness:.1f}'
                contamination_as_str = f'{contamination:.1f}'
                code = int(checkm_result["Translation table"])
        except KeyError:
            pass
        try:
            gtdbtk_classification = gtdbtk_results[old_name]
            if not gtdbtk_classification:
                gtdbtk_classification = gtdbtk_results[new_name]
        except KeyError:
            pass
        link_target = Path(new_name, 'index.html')
        new_name = f'<a href="{link_target}">{new_name}</a>'
        genomes_html += '<tr>\n<td>{}</td><td>{}</td><td>{:.1f}</td><td>{}</td><td>{}</td><td {}>{}</td>' \
                        '<td {}>{}</td><td>{}</td>\n</tr>'.format(old_name,
                                                                       new_name,
                                                                       genome_size,
                                                                       N50,
                                                                       code,
                                                                       colors[int(max((completeness-76)/5,0))],
                                                                       completeness_as_str,
                                                                       colors[int(max((24-contamination)/5, 0))],
                                                                       contamination_as_str,
                                                                       gtdbtk_classification)
    html = html.replace('GENOMES', genomes_html)
    return html


def _make_html_template() -> str:
    """Creates and returns the html base for injecting the content in."""
    return '''<!DOCTYPE html>
<html>
<head>
    <meta charset="utf-8">
    <title>all annotated genomes</title>
</head>
<body>

<script src="https://code.jquery.com/jquery-3.6.0.min.js" integrity="sha256-/xUj+3OJU5yExlq6GSYGSHk7tPXikynS7ogEvDej/m4=" crossorigin="anonymous"></script>

<link rel="stylesheet" type="text/css" href="https://cdn.datatables.net/1.11.5/css/jquery.dataTables.css">

<script type="text/javascript" charset="utf8" src="https://cdn.datatables.net/1.11.5/js/jquery.dataTables.js"></script>

<script type="text/javascript" charset="utf8">
$(document).ready( function () {
    $('#table_id').DataTable({
        "lengthMenu": [[20, 100, -1], [20, 100, "All"] ],
        });
} );
</script>

<style>
  th {
    background-color: white;
     }
  #f {
    font-family: Calibri, sans-serif;
    text-align: center;
    padding: 0px;
    margin: 0px;
     }
  #cg {
    color: green;
     }
  #cr {
    color: red;
     }
  #cb {
    color: blue;
     }
  #co {
    color: orange;
     }
  #cw {
    color: white;
     }
  #al {
    text-align: left;
      }

</style>

<div id=f>
<table id="table_id" class="display">
    <thead>
      <tr>
HEADERS
      </tr><
</thead>
<tbody>
GENOMES
</tbody>
</table> 
</div>
<div id=f>
<iframe src="" title="gene details" name="gene_details" style="border:none;width:100%;height:1000px;"></iframe>
</div>
</body>
</html>'''



import re
import pandas as pd
from pathlib import Path

from metaerg import context
from metaerg.datatypes import sqlite
from metaerg.datatypes.blast import BlastResult, BlastHit

DATAFRAME_INDEX = None
SUBSYSTEM_DATA = pd.DataFrame()
CONFIG_DATA_FILE_EXTENSION = '.config.txt'


def parse_functional_gene_config_file(file: Path, subsystems: list, index_data: list):
    gene_count = 0
    subsystem_count = 0
    current_subsystem = None
    with open(file) as handle:
        for line in handle:
            line = line.strip()
            if line.startswith("#") or not len(line):
                continue
            elif line.startswith(">"):
                current_subsystem = line[1:]
                subsystem_count += 1
            elif current_subsystem is not None:
                si = line.find(' ')
                subsystems.append({'profiles': line[:si], 'genes': ''})
                index_data.append((current_subsystem, line[si + 1:]))
                gene_count += 1
    context.log(f'Parsed {subsystem_count} subsystems, {gene_count} genes from {file} ...')


def init_functional_gene_config():
    context.log('Parsing functional gene config files...')
    global DATAFRAME_INDEX, SUBSYSTEM_DATA
    subsystems = []
    index_data = []
    base_config_file = Path(__file__).parent / 'run_and_read' / 'data' / 'functional_gene_data'
    parse_functional_gene_config_file(base_config_file, subsystems, index_data)
    user_config_dir = context.DATABASE_DIR / 'hmm' / 'user_config'
    if user_config_dir.exists() and user_config_dir.is_dir():
        for config_file in (context.DATABASE_DIR / 'user_config').glob('*' + CONFIG_DATA_FILE_EXTENSION):
            parse_functional_gene_config_file(user_config_dir / config_file.name, subsystems, index_data)
    DATAFRAME_INDEX = pd.MultiIndex.from_tuples(index_data, names=['subsystem', 'function'])
    SUBSYSTEM_DATA = pd.DataFrame(subsystems, dtype=str, index=DATAFRAME_INDEX)
    context.log(f'Complete configuration contains {len(subsystems)} genes.')


def match(blast_result: BlastResult) -> str:
    subsystems = ''
    i = 0
    for h in blast_result.hits:
        i += 1
        if i > 5:
            break
        if new_subsystem := match_hit(h):
            subsystems += ' ' + new_subsystem
    return subsystems


def match_hit(blast_hit: BlastHit, confidence=1) -> str:
    matching_subsystems = set()
    for sf in SUBSYSTEM_DATA.itertuples():
        for profile in sf.profiles.split('|'):
            if profile == blast_hit.hit.accession and blast_hit.aligned_length >= 0.7 * blast_hit.hit.length:
                matching_subsystems.add(f'[{sf.Index[0]}|{sf.Index[1]}|{confidence:.1f}]')
    return ' '.join(matching_subsystems).strip()


def cleanup_subsystem_str(subsystem_str: str) -> str:
    subsystem_str = subsystem_str.strip()
    collected_subsystems = {}
    for s in re.split(r'\s(?=\[)', subsystem_str):  # split at space if followed by [
        s = s.strip()[1:-1]
        if '|' in s:
            try:
                gene_subsystem, gene_function, confidence = s.split('|')
                confidence = float(confidence)
                hash_str = f'{gene_subsystem}|{gene_function}'
                prev_confidence = collected_subsystems[hash_str]
                if confidence > prev_confidence:
                    collected_subsystems[hash_str] = confidence
            except KeyError:
                collected_subsystems[hash_str] = confidence
            except ValueError:
                context.log(f'Error cleaning up subsystem "{s}"')

        else:
            collected_subsystems[s] = 0
    subsystem_str = ' '.join((f'[{val[0]}|{val[1]:.1f}]' for val in sorted(collected_subsystems.items(),
                                                                           key=lambda x: (-x[1], x[0]))))
    return subsystem_str.strip()


def aggregate(db_connection):
    aggregated_subsystem_data = SUBSYSTEM_DATA.copy().sort_index()
    unstructured_subsystems = {}
    for feature in sqlite.read_all_features(db_connection):
        if not len(feature.subsystems):
            continue
        for s in re.split(r'\s(?=\[)', feature.subsystems):  # split at space if followed b y [
            s = s[1:-1]
            try:
                gene_subsystem, gene_function, confidence = s.split('|')
                try:
                    aggregated_subsystem_data.loc[(gene_subsystem, gene_function), 'genes'] += \
                        f' {feature.id}@{confidence}'
                except TypeError:
                    aggregated_subsystem_data.loc[(gene_subsystem, gene_function), 'genes'] = feature.id
            except ValueError:
                gene_subsystems = s.split('|')
                try:
                    unstructured_subsystems[gene_subsystems[0]] += f' {feature.id}'
                except KeyError:
                    unstructured_subsystems[gene_subsystems[0]] = feature.id
    new_dataframe_rows = [{'profiles': '', 'genes': v} for k, v in unstructured_subsystems.items()]
    new_dataframe_index = [(k, k) for k, v in unstructured_subsystems.items()]
    new_dataframe_index = pd.MultiIndex.from_tuples(new_dataframe_index, names=['subsystem', 'function'])
    new_dataframe_data = pd.DataFrame(new_dataframe_rows, dtype=str, index=new_dataframe_index)
    aggregated_subsystem_data = pd.concat([aggregated_subsystem_data, new_dataframe_data])
    # print (aggregated_subsystem_data)
    return aggregated_subsystem_data



from pathlib import Path
from virtualenv import cli_run
import os
from shutil import which


def install_all_helper_programs(bin_dir: Path, path_to_signalp: Path, path_to_tmhmm: Path):
    # check for required programs
    success = True
    for cmd in 'git make gcc tar wget perl sed'.split():
        if cmd_loc := which(cmd):
            print(f'{cmd}: will use {cmd_loc}')
        else:
            print(f'{cmd}: not found. Please install first using your system\'s package manager.')
    if not success:
        print('aborted installation of helper programs.')
        return

    # Create the profile file.
    # To "activate" your metaerg installation, you will need to run, for example:
    # >source /home/my_name/bin/metaerg/bin/profile
    # (if that is the path to your installation)
    profile = f'''
    export BIOINF_PREFIX={bin_dir}
    PATH=$BIOINF_PREFIX/infernal/binaries:$PATH
    PATH=$BIOINF_PREFIX/hmmer2/src:$PATH
    PATH=$BIOINF_PREFIX/tmhmm/bin:$PATH
    PATH=$BIOINF_PREFIX/repeatscout:$PATH
    PATH=$BIOINF_PREFIX/repeatmasker:$PATH
    PATH=$BIOINF_PREFIX/minced:$PATH
    PATH=$BIOINF_PREFIX/hmmer3/bin:$PATH
    PATH=$BIOINF_PREFIX/ncbi-blast/bin:$PATH
    PATH=$BIOINF_PREFIX:$PATH
    export PATH
    '''
    profile_file = bin_dir / 'profile'
    with open(profile_file, "w") as profile_handle:
        profile_handle.write(profile)
    os.system(f'cat {profile_file}')
    for line in profile.split():
        if line.startswith('PATH='):
            os.environ["PATH"] += ':' + line.replace('PATH=$PATH:$BIOINF_PREFIX', str(bin_dir))
    print(os.environ["PATH"])
    os.chdir(bin_dir)
    # (minced) minced 0.4.2 https://github.com/ctSkennerton/minced
    os.system('wget -q https://github.com/ctSkennerton/minced/archive/refs/tags/0.4.2.tar.gz')
    os.system('tar -xf 0.4.2.tar.gz')
    os.chdir('minced-0.4.2')
    os.system('make')
    os.chdir(bin_dir)
    os.system('mv minced-0.4.2 minced')
    os.system('rm 0.4.2.tar.gz')
    # (aragorn) aragorn 1.2.41 https://www.ansikte.se/ARAGORN/Downloads/
    os.system('wget -q https://www.ansikte.se/ARAGORN/Downloads/aragorn1.2.41.c')
    os.system('gcc -O3 -ffast-math -finline-functions -o aragorn aragorn1.2.41.c')
    os.system('rm aragorn1.2.41.c')
    # (infernal) cmsearch 1.1.4 http://eddylab.org/infernal/
    os.system('wget -q http://eddylab.org/infernal/infernal-1.1.4-linux-intel-gcc.tar.gz')
    os.system('tar -xf infernal-1.1.4-linux-intel-gcc.tar.gz')
    os.system('mv infernal-1.1.4-linux-intel-gcc infernal')
    os.system('rm infernal-1.1.4-linux-intel-gcc.tar.gz')
    # (genometools) gt 1.6.2 http://genometools.org/tools/gt_ltrharvest.html
    os.system('wget -q http://genometools.org/pub/binary_distributions/gt-1.6.2-Linux_x86_64-64bit-barebone.tar.gz')
    os.system('tar -xf gt-1.6.2-Linux_x86_64-64bit-barebone.tar.gz')
    os.system('mv gt-1.6.2-Linux_x86_64-64bit-barebone genometools')
    os.system('ln -sf genometools/bin/gt .')
    os.system('rm gt-1.6.2-Linux_x86_64-64bit-barebone.tar.gz')
    # (tandem-repeat-finder) trf 4.09 https://tandem.bu.edu/trf/trf.html
    os.system('wget -q https://github.com/Benson-Genomics-Lab/TRF/releases/download/v4.09.1/trf409.linux64')
    os.system('chmod a+x trf409.linux64')
    os.system('ln -sf trf409.linux64 trf')
    # (RepeatScout) RepeatScout 1.0.5 https://github.com/mmcco/RepeatScout
    os.system('git clone https://github.com/mmcco/RepeatScout.git')
    os.system('mv RepeatScout repeatscout')
    os.chdir("repeatscout")
    os.system('make')
    os.chdir(bin_dir)
    #(nseg) nseg 1.0.1 https://github.com/jebrosen/nseg
    #(nseg is needed by repeatscout/repeatmasker)
    os.chdir(bin_dir)
    os.system('wget https://github.com/jebrosen/nseg/archive/refs/tags/v1.0.1.tar.gz')
    os.system('tar -xf v1.0.1.tar.gz')
    os.chdir("nseg-1.0.1")
    os.system('make')
    os.chdir(bin_dir)
    os.system('ln -sf nseg-1.0.1/nseg nseg')
    os.system('ln -sf nseg-1.0.1/nmerge nmerge')
    os.system('rm v1.0.1.tar.gz')
    # (rmblast) rmblastn 2.11.0 http://www.repeatmasker.org/RMBlast.html
    # (rmblast is needed by repeatmasker)
    os.system('wget -q http://www.repeatmasker.org/rmblast-2.11.0+-x64-linux.tar.gz')
    os.system('tar -xf rmblast-2.11.0+-x64-linux.tar.gz')
    os.system('ln -sf rmblast-2.11.0 rmblast')
    os.system('rm rmblast-2.11.0+-x64-linux.tar.gz')
    # (RepeatMasker) RepeatMasker 4.1.3 http://www.repeatmasker.org/RepeatMasker/
    os.system('wget -q http://www.repeatmasker.org/RepeatMasker/RepeatMasker-4.1.3.tar.gz')
    os.system('tar -xf RepeatMasker-4.1.3.tar.gz')
    os.system('mv RepeatMasker repeatmasker')
    os.chdir("repeatmasker")
    os.system(f'perl ./configure -default_search_engine rmblast -libdir {bin_dir / "repeatmasker" / "Libraries"} '
              f'-rmblast_dir {bin_dir / "rmblast" / "bin"} -trf_prgm {bin_dir / "trf"}')
    os.chdir(bin_dir)
    os.system('rm RepeatMasker-4.1.3.tar.gz')
    # (prodigal) prodigal 2.6.3 https://github.com/hyattpd/Prodigal
    os.system('wget -q https://github.com/hyattpd/Prodigal/releases/download/v2.6.3/prodigal.linux')
    os.system('chmod a+x prodigal.linux')
    os.system('ln -sf prodigal.linux prodigal')
    # (diamond) diamond 2.0.15 https://github.com/bbuchfink/diamond
    os.system('wget -q https://github.com/bbuchfink/diamond/releases/download/v2.0.15/diamond-linux64.tar.gz')
    os.system('tar -xf diamond-linux64.tar.gz')
    os.system('rm diamond-linux64.tar.gz')
    # (ncbi-blast) blastn 2.13.0 https://ftp.ncbi.nlm.nih.gov/blast/executables/blast+/LATEST/
    os.system('wget -q https://ftp.ncbi.nlm.nih.gov/blast/executables/blast+/LATEST/ncbi-blast-2.13.0+-x64-linux.tar.gz')
    os.system('tar -xf ncbi-blast-2.13.0+-x64-linux.tar.gz')
    os.system('mv ncbi-blast-2.13.0+ ncbi-blast')
    os.system('rm ncbi-blast-2.13.0+-x64-linux.tar.gz')
    # (hmmer-3) hmmsearch 3.3.2 http://hmmer.org
    os.system('wget -q http://eddylab.org/software/hmmer/hmmer-3.3.2.tar.gz')
    os.system('tar -xf hmmer-3.3.2.tar.gz')
    os.chdir('hmmer-3.3.2')
    os.system(f'./configure --prefix={bin_dir / "hmmer-3.3.2"}')
    os.system('make')
    os.system('make install')
    os.chdir('easel')
    os.system('make install')
    os.chdir(bin_dir)
    os.system('ln -s hmmer-3.3.2 hmmer3')
    # (hmmer-2) hmmsearch2 2.3.2 http://hmmer.org/
    # (required by antismash)
    os.system('wget -q http://eddylab.org/software/hmmer/hmmer-2.3.2.tar.gz')
    os.system('tar -xf hmmer-2.3.2.tar.gz')
    os.chdir('hmmer-2.3.2')
    os.system('./configure')
    os.system('make')
    os.chdir('src')
    os.system('mv hmmalign hmmalign2')
    os.system('mv hmmbuild hmmbuild2')
    os.system('mv hmmcalibrate hmmcalibrate2')
    os.system('mv hmmconvert hmmconvert2')
    os.system('mv hmmemit hmmemit2')
    os.system('mv hmmfetch hmmfetch2')
    os.system('mv hmmindex hmmindex2')
    os.system('mv hmmpfam hmmpfam2')
    os.system('mv hmmsearch hmmsearch2')
    os.chdir(bin_dir)
    os.system('ln -s hmmer-2.3.2 hmmer2')
    os.system('rm hmmer-2.3.2.tar.gz')
    #(muscle) muscle 3.8.31 https://drive5.com/muscle
    #(required by antismash)
    os.chdir(bin_dir)
    os.system('wget -q https://drive5.com/muscle/downloads3.8.31/muscle3.8.31_i86linux64.tar.gz')
    os.system('tar -xf muscle3.8.31_i86linux64.tar.gz')
    os.system('ln -sf muscle3.8.31_i86linux64 muscle')
    os.system('rm muscle3.8.31_i86linux64.tar.gz')
    #(fasttree) fasttree 2.1.11 https://microbesonline.org/fasttree
    # (required by antismash)
    os.system('wget -q https://microbesonline.org/fasttree/FastTreeMP')
    os.system('chmod a+x FastTreeMP')
    os.system('ln -sf FastTreeMP fasttree')
    os.system('ln -sf FastTreeMP FastTree')
    # (tmhmm) tmhmm 2.0c https://services.healthtech.dtu.dk/software.php
    if path_to_tmhmm:
        os.chdir(bin_dir)
        os.system(f'cp {path_to_tmhmm} {bin_dir}')
        os.system(f'tar -xf {path_to_tmhmm.name}')
        perl_exec = which('perl')  # \/usr\/bin\/perl on archlinux
        print(f'detected perl at {perl_exec}')
        os.system(f'sed -i "s|/usr/local/bin/perl|{perl_exec}|g" tmhmm-2.0c/bin/tmhmm')
        os.system(f'sed -i "s|/usr/local/bin/perl|{perl_exec}|g" tmhmm-2.0c/bin/tmhmmformat.pl')
        os.system('chmod a+x tmhmm-2.0c/bin/tmhmm')
        os.system('rm tmhmm-2.0c.Linux.tar.gz')
        os.system('ln -s tmhmm-2.0c tmhmm')
    # (signalp)
    if path_to_signalp:
        model_dir = Path(
            which('python')).parent.parent / 'lib' / 'python3.10' / 'site-packages' / 'signalp' / 'model_weights'
        print(f'signalp models will be installed at {model_dir}')
        os.chdir(bin_dir)
        os.system(f'cp {path_to_signalp} .')
        os.system(f'tar -xf {path_to_signalp.name}')
        os.system(f'{Path(which("python")).parent / "pip"} install signalp6_fast/signalp-6-package')
        os.system(f'cp -r signalp6_fast/signalp-6-package/models/* {model_dir}')
        os.system(f'rm -rf {path_to_signalp.name}')
    # (antismash) https://antismash.secondarymetabolites.org
    os.chdir(bin_dir)
    antismash_bin =  bin_dir / 'antismash-env' / 'bin'
    antismash_wrapper = bin_dir / 'antismash'
    with open(antismash_wrapper, "w") as handle:
        handle.write(f'#!/bin/sh\n{antismash_bin / "python"} {antismash_bin / "antismash"} "$@"\n')
    os.system('chmod a+x antismash')
    antismash_wrapper = bin_dir / 'download-antismash-databases'
    with open(antismash_wrapper, "w") as handle:
        handle.write(f'#!/bin/sh\n{antismash_bin / "python"} {antismash_bin / "download-antismash-databases"} "$@"\n')
    os.system('chmod a+x download-antismash-databases')
    cli_run(["antismash-env"])  # !python -m virtualenv antismash-env
    os.chdir("antismash-env")
    os.system('wget -q https://dl.secondarymetabolites.org/releases/6.1.1/antismash-6.1.1.tar.gz')
    os.system('tar -xf antismash-6.1.1.tar.gz')
    os.system('rm antismash-6.1.1.tar.gz')
    os.system('./bin/pip install --upgrade ./antismash-6.1.1')



import os
import gzip
import shutil
from math import log10
from pathlib import Path
from concurrent import futures

from metaerg.datatypes.blast import DBentry, TabularBlastParser
from metaerg.datatypes import sqlite
from metaerg import context
from metaerg import functional_gene_configuration


def _run_programs(genome_name, contig_dict, db_connection, result_files):
    cds_aa_file = context.spawn_file('cds.faa', genome_name)
    hmm_db = context.DATABASE_DIR / 'hmm' / 'functional_genes.hmm'
    # the domtbl format includes alignment starts and ends, so we can use that information...
    context.run_external(f'hmmscan -o /dev/null -E 1e-6 --domtblout {result_files[0]} {hmm_db} {cds_aa_file}')


def _read_results(genome_name, contig_dict, db_connection, result_files) -> int:
    db_entry = None
    hmm_db = context.DATABASE_DIR / 'hmm' / 'functional_genes.hmm'
    functional_gene_db = {}
    with open(hmm_db) as handle:
        for line in handle:
            if line.startswith('NAME'):
                name = line.split()[1]
                db_entry = DBentry(domain='functional_genes', descr='', accession=name, gene=name)
                functional_gene_db[name] = db_entry
            elif line.startswith('ACC'):
                db_entry.accession = line[4:].strip()
            elif line.startswith('NC'):
                db_entry.min_score = float(line[4:].strip().split()[0])
            elif line.startswith('TC'):
                db_entry.min_t_score = float(line[4:].strip().split()[0])
            elif line.startswith('DESC'):
                db_entry.descr = line[4:].strip()
            elif line.startswith('LENG'):
                db_entry.length = int(line[4:].strip())
    context.log(f'({genome_name}) {len(functional_gene_db)} functional gene profiles in database.')

    def get_db_entry(db_id) -> DBentry:
        return functional_gene_db[db_id]

    with TabularBlastParser(result_files[0], 'HMMSCAN_DOM_TABLE', get_db_entry) as handle:
        hit_count = 0
        for blast_result in handle:
            feature = sqlite.read_feature_by_id(db_connection, blast_result.query())
            if not feature:
                raise Exception(f'Found results for unknown feature {blast_result.query()}, '
                                f'may need to rerun metaerg with --force')
            blast_result.hits = blast_result.hits[:10]
            feature.hmm = blast_result
            for h in blast_result.hits:
                db_entry = h.hit
                confidence = 1.0
                if h.aligned_length / db_entry.length < 0.7:
                    continue
                if db_entry.min_score:
                    if h.score < db_entry.min_score:
                        continue
                    confidence = h.score / db_entry.min_t_score
                else:
                    if h.evalue > 0:
                        confidence = min(1.0, - log10(h.evalue) / 100)
                if new_subsystem := functional_gene_configuration.match_hit(h, confidence):
                    hit_count += 1
                    feature.subsystems = functional_gene_configuration.cleanup_subsystem_str(
                        feature.subsystems.strip() + ' ' + new_subsystem)
                break
            sqlite.update_feature_in_db(db_connection, feature)
    return hit_count


@context.register_annotator
def run_and_read_canthyd():
    return ({'pipeline_position': 101,
             'annotator_key': 'hmm',
             'purpose': 'identification of functional genes with HMMs',
             'programs': ('hmmscan',),
             'databases': (Path('hmm', 'functional_genes.hmm'),),
             'result_files': ('hmm',),
             'run': _run_programs,
             'read': _read_results})

FUNCTIONAL_GENE_URLS = ('https://zenodo.org/record/6365663/files/carbon.cycle.sub.hmm',
                        'https://zenodo.org/record/6365663/files/c1.cycle.sub.hmm',
                        'https://zenodo.org/record/6365663/files/methane.cycle.sub.hmm',
                        'https://zenodo.org/record/6365663/files/misc.cycle.sub.hmm',
                        'https://zenodo.org/record/6365663/files/nitro.cycle.sub.hmm',
                        'https://zenodo.org/record/6365663/files/oxygen.cycle.sub.hmm',
                        'https://zenodo.org/record/6365663/files/sulphur.cycle.sub.hmm',
                        'https://zenodo.org/record/6365663/files/hydrogenase.hmm',
                        'https://github.com/dgittins/CANT-HYD-HydrocarbonBiodegradation/raw/main/HMMs/concatenated%20HMMs/CANT-HYD.hmm')


def add_hmm_file_to_db(hmm_file: Path, db_handle, names_done: set, targets: set):
    with open(hmm_file) as hmm_reader:
        duplicates = []
        written = 0
        read = 0
        hmm_lines = []
        for line in hmm_reader:
            if line.startswith('HMMER3/f'):
                hmm_lines.clear()
                read += 1
                name = ''
            hmm_lines.append(line)
            if line.startswith('//'):
                if name in names_done:  # avoid duplicates
                    duplicates.append(name)
                    continue
                names_done.add(name)
                if name in targets:
                    written += 1
                    for l in hmm_lines:
                        if l.startswith('ACC'):
                            pass
                        else:
                            db_handle.write(l)
                elif hmm_file.name == 'CANT-HYD.hmm':
                    print(''.join(hmm_lines[:20]))
            elif line.startswith('NAME'):
                name = line[4:].strip()
    print(f'{hmm_file.name}: read {read}, written {written}, duplicates {len(duplicates)}.')


@context.register_database_installer
def install_functional_gene_databases():
    if 'S' not in context.TASKS:
        return

    functional_gene_configuration.init_functional_gene_config()
    all_target_hmm_names = set()
    for t in functional_gene_configuration.SUBSYSTEM_DATA['profiles']:
        for t1 in t.split('|'):
            all_target_hmm_names.add(t1)

    hmm_dir = context.DATABASE_DIR / 'hmm'
    hmm_dir.mkdir(exist_ok=True, parents=True)
    user_hmm_dir = hmm_dir / 'user_hmm'
    (hmm_dir / 'user_config').mkdir(exist_ok=True, parents=True)  # used by functional_gene_configuration.py
    user_hmm_dir.mkdir(exist_ok=True, parents=True)  # used below
    context.log(f'Installing functional gene hmm database at {hmm_dir}...')

    current_working_dir = os.getcwd()
    # fetch bd type oxygen reductases
    bdor_dir = hmm_dir / 'bdor'
    if bdor_dir.exists():
        shutil.rmtree(bdor_dir)
    bdor_dir.mkdir()
    os.chdir(bdor_dir)
    os.system('git init')
    os.system('git remote add bdor https://github.com/ranjani-m/cytbd-superfamily.git')
    os.system('git fetch bdor')
    os.system('git checkout bdor/main -- Cytbd_HMM_2020_paper_version')
    with open(hmm_dir / 'bdor.hmm', 'w') as writer:
        for hmm_file in sorted((bdor_dir / 'Cytbd_HMM_2020_paper_version').glob('*.hmm')):
            os.system(f'sed -i "s|^NAME.*|NAME  bdor_{hmm_file.stem}|" {hmm_file}')
            context.run_external(f'/bio/bin/hmmer3/bin/hmmconvert {hmm_file}', stdout=writer)
    # fetch heme copper oxidase hmms
    hco_dir = hmm_dir / 'hco'
    if hco_dir.exists():
        shutil.rmtree(hco_dir)
    hco_dir.mkdir()
    os.chdir(hco_dir)
    os.system('git init')
    os.system('git remote add hco https://github.com/ranjani-m/HCO.git')
    os.system('git fetch hco')
    os.system('git checkout hco/main -- HMM')
    with open(hmm_dir / 'hco.hmm', 'w') as writer:
        for hmm_file in sorted((hco_dir / 'HMM').glob('*.hmm')):
            os.system(f'sed -i "s|^NAME.*|NAME  hco_{hmm_file.stem}|" {hmm_file}')
            context.run_external(f'hmmconvert {hmm_file}', stdout=writer)

    # unzip hmms from data
    hmm_data_dir = Path(__file__).parent / 'data'
    print (hmm_data_dir)
    for hmm_file in hmm_data_dir.glob('*.hmm.gz'):
        with gzip.open(hmm_file, 'rb') as f_in:
            dest_hmm_file = hmm_dir / hmm_file.stem
            print ('unzipping to ', dest_hmm_file)
            with open(dest_hmm_file, 'wb') as f_out:
                shutil.copyfileobj(f_in, f_out)

    with futures.ThreadPoolExecutor() as executor:
        outcomes = []
        for url in FUNCTIONAL_GENE_URLS:
            destination_file = hmm_dir / Path(url).name
            if context.FORCE or not destination_file.exists() or not destination_file.stat().st_size:
                outcomes.append(executor.submit(context.download, url, destination_file))
        for future in futures.as_completed(outcomes):
            future.result()

    context.log('Checking data sanity...')
    all_hmms_file = hmm_dir / 'functional_genes.hmm'
    with open(all_hmms_file, 'w') as output:
        names_done = set()
        for f in hmm_dir.glob('*.hmm'):
            hmm_file = f.absolute()
            if f == all_hmms_file:
                continue
            add_hmm_file_to_db(hmm_file, output, names_done, all_target_hmm_names)
        for f in user_hmm_dir.glob('*.hmm'):
            hmm_file = f.absolute()
            add_hmm_file_to_db(hmm_file, output, names_done, all_target_hmm_names)
    context.run_external(f'hmmpress -f {hmm_dir / "functional_genes.hmm"}')
    os.chdir(current_working_dir)



import pandas as pd

from metaerg import context
from metaerg.datatypes import fasta
from metaerg.datatypes import sqlite

def _run_programs(genome_name, contig_dict, feature_data: pd.DataFrame, result_files):
    pass


def _read_results(genome_name, contig_dict, db_connection, result_files) -> int:
    j = 0
    for feature in sqlite.read_all_features(db_connection):
        if context.RENAME_CONTIGS:
            # contigs already contain genome name
            feature.id = context.DELIMITER.join((feature.contig, f'{j:05d}'))
        else:
            feature.id = context.DELIMITER.join((genome_name, feature.contig, f'{j:05d}'))
        j += 1
        sqlite.update_feature_in_db(db_connection, feature)

    cds_count = sum(1 for f in sqlite.read_all_features(db_connection, type='CDS'))
    rna_count = sum(1 for f in sqlite.read_all_features(db_connection, type=sqlite.RNA_TARGETS))

    cds_file = context.spawn_file('cds.faa', genome_name)
    context.log(f'({genome_name}) Now writing {cds_count} proteins to fasta at {cds_file}...')
    fasta.write_features_to_fasta(db_connection, 'aa', cds_file, targets=('CDS',))
    rna_file = context.spawn_file('rna.fna', genome_name)
    context.log(f'({genome_name}) Now writing {rna_count} RNA genes and features to fasta at {rna_file}...')
    fasta.write_features_to_fasta(db_connection, 'nt', rna_file, targets=sqlite.RNA_TARGETS)
    return j


@context.register_annotator
def run_and_read_trf():
    return ({'pipeline_position': 66,
             'annotator_key': 'write_genes',
             'purpose': 'feature ID generation',
             'programs': (),
             'result_files': (),
             'run': _run_programs,
             'read': _read_results})



from pathlib import Path
from math import log10

from metaerg.datatypes import sqlite
from metaerg.datatypes.blast import DBentry, BlastHit, BlastResult, taxon_at_genus
from metaerg import context

MAX_BLAST_HITS = 10
COLORS = 'id=cr id=cr id=co id=cb id=cg'.split()


@context.register_html_writer
def write_html(genome_name, db_connection, genome_properties:dict, dir):
    """Writes a html file for each feature to dir <file>"""
    dir = Path(dir, genome_name)
    dir.mkdir(exist_ok=True, parents=True)
    file = dir / 'feature-details.html'

    feature_html = ''
    for feature in sqlite.read_all_features(db_connection, type=('CDS', 'rRNA', 'ncRNA', 'retrotransposon')):
        if feature.type in ('CDS', 'rRNA', 'ncRNA', 'retrotransposon'):
            feature_html += make_feature_html(feature, genome_properties['dominant taxon'])

    html = _make_html_template().replace('FEATURE_HTML', feature_html)
    with open(file, 'w') as handle:
        handle.write(html)


def make_blast_table_html(blast_result: BlastResult, f_length, dominant_taxon, include_id, title: str,
                          headers: str ='percent id|query align|hit align|description|taxon') -> str:
    if blast_result:
        html = f'<h3>{title}</h3>\n'
        html += '<table><thead><tr>\n'
        columns = headers.split('|')
        for column in columns:
            if column == 'description':
                html += f'<th id=al>{column}</th>\n'
            else:
                html += f'<th>{column}</th>\n'
        html += '</tr><thead>\n<tbody>\n'
        for h in blast_result.hits[:MAX_BLAST_HITS]:
            if 'evalue' == columns[0]:
                percent_id = f'{h.evalue:.1e}'
                color_value = len(COLORS)-1 if not h.evalue else min(0, max(int(-log10(h.evalue)/20),len(COLORS)-1))
                percent_id_color = COLORS[color_value]
            else:
                percent_id = int(h.percent_id) if h.percent_id else ''
                percent_id_color = COLORS[min(int(h.percent_id/20), len(COLORS)-1)]
            if include_id:
                descr = h.hit.accession + ' ' + h.hit.descr
            else:
                descr = h.hit.descr
            html += '''<tr>
            <td {}>{}</td>
            <td {}>{}</td>
            <td {}>{}</td>
            <td id=al>{}</td>
            <td {}>{}</td>
            </tr>'''.format(
                percent_id_color,
                percent_id,

                COLORS[min(int(100 * h.aligned_length/f_length / 20), len(COLORS)-1)],
                int(100 * min(1.0, h.aligned_length/f_length)),

                COLORS[min(int(100 * h.aligned_length/h.hit.length / 20), len(COLORS)-1)],
                int(100 * min(1.0, h.aligned_length/h.hit.length)),

                descr,

                COLORS[int(len(COLORS) * len(set(h.hit.taxon.split()) & set(dominant_taxon.split()))
                           / (len (h.hit.taxon.split()) + 1))],
                taxon_at_genus(h.hit.taxon)
            )
        html += '</tbody>\n'
        html += '</table>\n'
        return html
    else:
        return ''


def make_feature_html(f, dominant_taxon) -> str:
    html = _make_feature_html_template()
    html = html.replace('FEATURE_ID', f.id)
    if f.taxon:
        html = html.replace('HEADER', f'>{f.id} {f.descr} [{taxon_at_genus(f.taxon)}]')
    else:
        html = html.replace('HEADER', f'>{f.id} {f.descr}')
    html = html.replace('SEQUENCE', f.aa_seq if 'CDS' == f.type else f.nt_seq)
    if f.type == 'CDS':
        length = (f.end - f.start) // 3
        length_unit = 'aa'
    else:
        length = f.end - f.start
        length_unit = 'nt'
    html = html.replace('BLAST_TABLE', make_blast_table_html(f.blast, length, dominant_taxon, include_id = False,
                                                             title='Top Blast Hits'))
    html = html.replace('CDD_TABLE', make_blast_table_html(f.cdd, length, dominant_taxon, include_id = True,
                                                           title='Top Conserved Domain Database Hits',
                                                           headers='percent id|query align|hit align|description| '))
    html = html.replace('HMM_TABLE', make_blast_table_html(f.hmm, length, dominant_taxon, include_id = True,
                                                           title='Top Functional Gene HMM Hits',
                                                           headers='evalue|query align|hit align|description| '))
    attribute_html = '<table>\n'
    f.tmh = '' if not f.tmh else f.tmh
    f_as_dict = {k:v for k,v in f}
    attribute_html += ''.join(f'<tr><td id=al>{k}</td><td id=al>{f_as_dict[k]}</td></tr>\n' for k in
                              ('start', 'end', 'strand', 'type', 'inference', 'subsystems', 'descr', 'taxon', 'notes',
                               'antismash', 'signal_peptide', 'tmh', 'tmh_topology'))
    attribute_html += f'<tr><td id=al>length</td><td id=al>{length} {length_unit}</td></tr>\n'
    attribute_html += '</table>\n'
    html = html.replace('ATTRIBUTE_TABLE', attribute_html)
    return html


def _make_html_template() -> str:
    return '''<!DOCTYPE html>
<html>
<head>
    <meta charset="utf-8">
    <title>Gene Details</title>
</head>
<body>
    <style>
        div {font-family: Calibri, sans-serif; word-wrap: break-word;}
        h3 {text-align: left; font-family: Calibri, sans-serif; margin: 10px 0px 0 px 0px;}
        #al {text-align: left;}
        #cg {color: green;}
        #cr {color: red;}
        #cb {color: blue;}
        #co {color: orange;}
        th {
          background-color: white; margin: 0px 10px 0 px 0px;
            }
        #al {
          text-align: left;
            }
        #f {
          font-family: Calibri, sans-serif;
          text-align: center;
          padding: 0px;
          margin: 0px;
           }
        #gn {
          background-color: gainsboro;
            }
    </style>
    <p id=cr>This is a very large file - it takes some time until all genes are ready for display.</p>
    FEATURE_HTML
</body>
</html>
'''


def _make_feature_html_template() -> str:
    return '''
    <div id=gn><h3 id="FEATURE_ID">HEADER</h3></div>
    <div>SEQUENCE</div>
    <div id=f>
    BLAST_TABLE
    CDD_TABLE
    HMM_TABLE
    <h3>Other attributes</h3>
    ATTRIBUTE_TABLE
    </div>
'''


from pathlib import Path

from metaerg import context
from metaerg.datatypes import sqlite
from metaerg.datatypes.blast import taxon_at_genus


@context.register_html_writer
def write_html(genome_name, db_connection, genome_properties:dict, dir):
    dir.mkdir(exist_ok=True, parents=True)
    file = Path(dir, genome_name, "feature_table.html")
    file.parent.mkdir(exist_ok=True, parents=True)
    with open(Path(file), 'w') as handle:
        handle.write(make_html(genome_name, db_connection, genome_properties))


def get_empty_format_dict():
    return {'f_id': '',
            'strand': '',
            'length': 0,
            'type': '',
            'destination': '',
            'subsystem': '',
            'has_cdd': '',
            'ident': '',
            'align': '',
            'recall': '',
            'description': '',
            'taxon': '',
            'ci': '', 'ca': '', 'cr': '', 'ct': ''}


def format_feature(f: sqlite.Feature, format_hash, dominant_taxon, colors):
    format_hash['f_id'] = f.id
    format_hash['taxon'] = taxon_at_genus(f.taxon)
    format_hash['type'] = f.type
    if f.type in ('CDS', 'rRNA', 'ncRNA', 'retrotransposon'):
        format_hash['description'] = '<a target="Gene Details" href="feature-details.html#{}">{}</a>'.format(
            f'{f.id}', f.descr)
    else:
        format_hash['description'] = f.descr
    if f.type in ('CDS', 'tRNA', 'rRNA', 'ncRNA', 'tmRNA', 'retrotransposon'):
        format_hash['strand'] = "+" if f.strand > 0 else "-"
    else:
        format_hash['strand'] = ''
    format_hash['length'] = len(f.aa_seq if 'CDS' == f.type else f.nt_seq)
    match f.tmh, f.signal_peptide, f.type:
        case [_, 'LIPO', _]:
            format_hash['destination'] = 'lipoprotein'
        case [1, _, _]:
            format_hash['destination'] = 'membrane anchor'
        case [tmh, _, _] if tmh > 1:
            format_hash['destination'] = 'membrane'
        case [_, sp, _] if len(sp):
            format_hash['destination'] = 'envelope'
        case [_, _, 'CDS']:
            format_hash['destination'] = 'cytoplasm'
        case [*_]:
            format_hash['destination'] = ''
    format_hash['subsystem'] = f.subsystems
    format_hash['has_cdd'] = 'Y' if f.cdd is not None else ''
    if f.blast:
        format_hash['ident'] = f'{f.blast.hits[0].percent_id:.1f}'
        format_hash['ci'] = colors[min(int(f.blast.hits[0].percent_id / 20), len(colors) - 1)]
        format_hash['align'] = f'{f.blast.percent_aligned():.1f}'
        format_hash['ca'] = colors[min(int(f.blast.percent_aligned() / 20), len(colors) - 1)]
        format_hash['recall'] = f'{f.blast.percent_recall():.1f}'
        format_hash['cr'] = colors[min(int(f.blast.percent_recall() / 20), len(colors) - 1)]
    dominant_taxon = dominant_taxon.split()
    taxon = f.taxon.split()
    format_hash['ct'] = colors[int(len(colors) * len(set(taxon) & set(dominant_taxon)) / (len(taxon) + 1))]


def format_hash_to_html(format_hash):
    return '''<tr>
    <td id=al>{f_id}</td> <td>{strand}</td> <td>{length:,}</td> <td>{type}</td> <td>{destination}</td> 
    <td>{subsystem}</td> <td>{has_cdd}</td> <td{ci}>{ident}</td> <td{ca}>{align}</td> <td{cr}>{recall}</td> 
    <td id=al>{description}</td>
    <td{ct}>{taxon}</td>
    </tr>'''.format(**format_hash)


def make_html(genome_name, db_connection, genome_properties:dict) -> str:
    """Injects the content into the html base, returns the html."""
    html = _make_html_template()
    html = html.replace('GENOME_NAME', genome_name)
    colors = [' id=cr', ' id=cr', ' id=co', ' id=cb', ' id=cg']

    # table header
    table_headers = ''
    for column in 'id strand length type location subsystem CDD ident align recall description taxon'.split():
        if column in 'id description':
            table_headers += f'<th id=al>{column}</th>\n'
        else:
            table_headers += f'<th>{column}</th>\n'
    html = html.replace('TABLE_HEADERS', table_headers)
    # table body
    table_body = ''
    previous_repeats = []
    prev_f = None
    for f in sqlite.read_all_features(db_connection):
        if f.type in ('crispr_repeat', 'repeat') and (not len(previous_repeats) or (f and f.type == prev_f.type)):
            previous_repeats.append(f)
        elif len(previous_repeats):
            format_hash = get_empty_format_dict()
            format_hash['description'] = previous_repeats[0].id + ' ... ' + previous_repeats[-1].id
            format_hash['type'] = f'[{len(previous_repeats)} {prev_f.type}s]' if len(previous_repeats) > 1 \
                                  else prev_f.type
            format_hash['length'] = previous_repeats[-1].end - previous_repeats[0].start
            previous_repeats.clear()
            table_body += format_hash_to_html(format_hash)
            if f.type in ('crispr_repeat', 'repeat'):
                previous_repeats.append(f)
            else:
                format_hash = get_empty_format_dict()
                format_feature(f, format_hash, genome_properties['dominant taxon'], colors)
                table_body += format_hash_to_html(format_hash)
        else:
            format_hash = get_empty_format_dict()
            format_feature(f, format_hash, genome_properties['dominant taxon'], colors)
            table_body += format_hash_to_html(format_hash)
        prev_f = f
    html = html.replace('TABLE_BODY', table_body)
    return html


def _make_html_template() -> str:
    """Creates and returns the html base for injecting the content in."""
    return '''<!DOCTYPE html>
<html>
<head>
    <meta charset="utf-8">
    <title>GENOME_NAME - all features</title>
</head>
<body>

<script src="https://code.jquery.com/jquery-3.6.0.min.js" integrity="sha256-/xUj+3OJU5yExlq6GSYGSHk7tPXikynS7ogEvDej/m4=" crossorigin="anonymous"></script>
<link rel="stylesheet" type="text/css" href="https://cdn.datatables.net/1.11.5/css/jquery.dataTables.css">
<script type="text/javascript" charset="utf8" src="https://cdn.datatables.net/1.11.5/js/jquery.dataTables.js"></script>
<script type="text/javascript" charset="utf8">
$(document).ready( function () {
    $('#table_id').DataTable({
        "lengthMenu": [[20, 100, -1], [20, 100, "All"] ],
        "bSort" : false
        });
} );
</script>

<style>
  th {
    background-color: white;
      }
  #f {
    font-family: Calibri, sans-serif;
    text-align: center;
    padding: 0px;
    margin: 0px;
     }
  #cg {
    color: green;
     }
  #cr {
    color: red;
     }
  #cb {
    color: blue;
     }
  #co {
    color: orange;
     }
  #cw {
    color: white;
     }
  #al {
    text-align: left;
      }
</style>

<div id=f>
<table id="table_id" class="display">
    <thead>
        <tr>
TABLE_HEADERS
        </tr>
    </thead>
    <tbody>
TABLE_BODY
    </tbody>
</table> 
</div>
<div id=f>
<iframe src="" title="gene details" name="gene_details" style="border:none;width:100%;height:1000px;"></iframe>
</div>
</body>
</html>'''



import re
import shutil
import subprocess
import time
import os
import traceback
from multiprocessing import cpu_count
from pathlib import Path

import httpx

from metaerg import registry

BASE_DIR = Path()
TEMP_DIR = Path()
HTML_DIR = Path()
DATABASE_DIR = Path()
CHECKM_DIR = Path()
GTDBTK_DIR = Path()
GENOME_NAME_MAPPING_FILE = ''
CONTIG_FILES = []
GENOME_NAMES = []
LOG_FILE = ''

FORCE = False
MULTI_MODE = False
RENAME_CONTIGS = False
RENAME_GENOMES = False
MIN_CONTIG_LENGTH = 0
TRANSLATION_TABLE = 0
ACTIVE_ANNOTATORS = set('antismash aragorn cdd cmscan diamond_and_blastn hmm write_genes ltr_harvest minced prodigal '
                        'signalp repeat_masker tmhmm trf'.split())
READ_ONLY = False
DELIMITER = ''
CPUS_PER_GENOME = 0
CPUS_AVAILABLE = 0
PARALLEL_ANNOTATIONS = 0
START_TIME = 0
LOG_TOPICS = set()
FILE_EXTENSION = ''
METAERG_MODE_RUN = 1
METAERG_MODE_DOWNLOAD_DATABASE = 2
METAERG_MODE_CREATE_DATABASE = 3
METAERG_MODE_INSTALL_DEPS = 4
METAERG_MODE = METAERG_MODE_RUN
TASKS = 'all'
PREFIX = 'g'
BIN_DIR =''
PATH_TO_SIGNALP = None
PATH_TO_TMHMM = None

def init(contig_file, database_dir, rename_contigs, rename_genomes, min_contig_length, cpus, force, file_extension,
         translation_table, delimiter, checkm_dir, gtdbtk_dir, prefix, create_database, download_database,
         install_deps, path_to_signalp, path_to_tmhmm, log_topics='', contig_mode = False, skip_step='', read_only=True,
         run_step=''):
    global BASE_DIR, TEMP_DIR, HTML_DIR, DATABASE_DIR, CHECKM_DIR, GTDBTK_DIR, GENOME_NAME_MAPPING_FILE, MULTI_MODE,\
           RENAME_CONTIGS, RENAME_GENOMES, MIN_CONTIG_LENGTH, FORCE, FILE_EXTENSION, TRANSLATION_TABLE, \
           CPUS_PER_GENOME, CPUS_AVAILABLE, START_TIME, LOG_TOPICS, PARALLEL_ANNOTATIONS, METAERG_MODE, \
           GENOME_NAMES, CONTIG_FILES, DELIMITER, LOG_FILE, TASKS, PREFIX, BIN_DIR, PATH_TO_SIGNALP, PATH_TO_TMHMM, \
           ACTIVE_ANNOTATORS, READ_ONLY

    START_TIME = time.monotonic()
    LOG_TOPICS = set(log_topics.split())
    LOG_FILE = Path('log.txt').absolute()
    log('Initializing execution environment with command line arguments...')

    if install_deps:
        METAERG_MODE = METAERG_MODE_INSTALL_DEPS
        BIN_DIR = Path(install_deps).absolute()
        if path_to_signalp and Path(path_to_signalp).is_file():
            PATH_TO_SIGNALP = Path(path_to_signalp).absolute()
        else:
            log('Warning: path to signalp tarbal not provided or wrong; signalp will not be installed.')
        if path_to_tmhmm and Path(path_to_tmhmm).is_file():
            PATH_TO_TMHMM = Path(path_to_tmhmm).absolute()
        else:
            log('Warning: path to tmhmm tarbal not provided or wrong; tmhmm will not be installed.')
        log(f'Ready to install helper programs at {BIN_DIR} with {PATH_TO_SIGNALP}, {PATH_TO_TMHMM}.')
        return

    DATABASE_DIR = Path(database_dir).absolute()
    if not DATABASE_DIR.is_dir():
        raise Exception("No database dir provided or database dir is not a dir.")
    if download_database:
        METAERG_MODE = METAERG_MODE_DOWNLOAD_DATABASE
        log(f'Ready to download databases.')
        return
    elif create_database:
        METAERG_MODE = METAERG_MODE_CREATE_DATABASE
        if create_database == 'all':
            TASKS = 'PVEBRCSA'
        else:
            TASKS = create_database
        log(f'Ready to create databases from scratch with tasks {TASKS}.')
        return
    else: # we're going to annotate genomes...
        # (1) collect contig files
        if ',' in contig_file:
            for f in contig_file.split(','):
                f = Path(f).absolute()
                if not f.exists():
                    log(f'Did not find contig file "{f}"')
                    continue
                CONTIG_FILES.append(f)
        else:
            contig_file = Path(contig_file).absolute()
            if contig_file.is_dir():
                CONTIG_FILES = [f.absolute() for f in sorted(contig_file.glob(f'*{FILE_EXTENSION}')) if f.is_file()]
            else:
                if contig_file.exists():
                    CONTIG_FILES = [contig_file]
        if len(CONTIG_FILES):
            BASE_DIR = CONTIG_FILES[0].parent
        else:
            log(f'No contig files to annotate, did you type the correct file or file extension?')
            exit(1)
        # (2) manage file structure / folders
        TEMP_DIR = BASE_DIR / 'temp'
        CHECKM_DIR = Path(checkm_dir).absolute()
        GTDBTK_DIR = Path(gtdbtk_dir).absolute()
        GENOME_NAME_MAPPING_FILE = TEMP_DIR / 'genome.name.mapping.txt'
        HTML_DIR = BASE_DIR / 'html'
        for folder in (TEMP_DIR, HTML_DIR):
            if folder.exists():
                print(f'Warning: may overwrite existing files in {folder}...')
                if folder.is_file():
                    print(f'Expected dir at {folder}, found regular file, crash! Delete this file first')
                    exit(1)
            else:
                folder.mkdir()
        if run_step:
            ACTIVE_ANNOTATORS = {step for step in run_step.split(',')}
        if contig_mode:
            TRANSLATION_TABLE = -1
            ACTIVE_ANNOTATORS.remove('repeat_masker')
        for skipped_step in skip_step.split(','):
            try:
                ACTIVE_ANNOTATORS.remove(skipped_step)
            except KeyError:
                pass
        READ_ONLY = read_only
        # (3) set some global variables
        RENAME_CONTIGS = rename_contigs
        RENAME_GENOMES = rename_genomes
        MIN_CONTIG_LENGTH = int(min_contig_length)
        FORCE = force
        FILE_EXTENSION = file_extension
        DELIMITER = delimiter
        PREFIX = prefix
        TRANSLATION_TABLE = translation_table
        # (4) set up multitreading
        CPUS_PER_GENOME = int(cpus)
        CPUS_AVAILABLE = cpu_count() // 2
        if CPUS_PER_GENOME > 0:
            CPUS_PER_GENOME = min(CPUS_PER_GENOME, CPUS_AVAILABLE)
        else:
            CPUS_PER_GENOME = CPUS_AVAILABLE
        PARALLEL_ANNOTATIONS = CPUS_PER_GENOME
        CPUS_PER_GENOME = max(1, CPUS_PER_GENOME // len(CONTIG_FILES))
        PARALLEL_ANNOTATIONS = int(PARALLEL_ANNOTATIONS / CPUS_PER_GENOME)
        if len(CONTIG_FILES) == 1:
            PARALLEL_ANNOTATIONS = 1
        log(f'Detected {CPUS_AVAILABLE} available threads/cpus, will use {CPUS_PER_GENOME} per genome with '
            f'{PARALLEL_ANNOTATIONS} genomes annotated in parallel.')
        MULTI_MODE = len(CONTIG_FILES) > 1
        if  (TEMP_DIR / 'metaerg_progress').exists and (TEMP_DIR / 'metaerg_progress').is_dir():
            log(f'Detected previous results - maintaining file structure for compatibility.')
            MULTI_MODE = True
        # (5) manage genome and contig naming
        if RENAME_GENOMES:
            log('Will create new names for genomes.')
            RENAME_CONTIGS = True
        else:
            log('Using original filenames as genome names.')
        if RENAME_CONTIGS:
            log('Will rename contigs.')
        else:
            log('Keeping original contig identifiers.')
        # (6) more about genome names and consistency between metaerg (re)runs
        if GENOME_NAME_MAPPING_FILE.exists():
            log(f'Detected previous results -  maintaining genome name mappings for consistency.')
            prev_name_mappings = {}
            index_for_next_genome = 0
            with open(GENOME_NAME_MAPPING_FILE) as reader:
                for line in reader:
                    words = line.split()
                    prev_name_mappings[words[1]] = words[0]
                    if match := re.search(r'\d+', words[0]):
                        index = int(match.group(0))
                        if index > index_for_next_genome:
                            index_for_next_genome = index
            GENOME_NAMES = []
            for contig_file in CONTIG_FILES:
                try:
                    GENOME_NAMES.append(prev_name_mappings[contig_file.stem])
                except KeyError:
                    if RENAME_GENOMES:
                        index_for_next_genome += 1
                        new_name = f'{PREFIX}{index_for_next_genome:0>4}'
                        GENOME_NAMES.append(new_name)
                        with open(GENOME_NAME_MAPPING_FILE, 'a') as mapping_file:
                            mapping_file.write(f'{new_name}\t{contig_file.stem}\t{contig_file}\n')
                            log(f'Detected new contig file {contig_file.name} - renamed to {new_name} and added to'
                                f'name mapping file.')
                    else:
                        GENOME_NAMES.append(contig_file.stem)
        else:
            if RENAME_GENOMES:
                GENOME_NAMES = [f'{PREFIX}{CONTIG_FILES.index(f):0>4}' for f in CONTIG_FILES]
            else:
                GENOME_NAMES = [f.stem for f in CONTIG_FILES]
            log(f'Writing genome names to {GENOME_NAME_MAPPING_FILE}...')
            with open(GENOME_NAME_MAPPING_FILE, 'w') as mapping_file:
                for n, o in zip(GENOME_NAMES, CONTIG_FILES):
                    mapping_file.write(f'{n}\t{o.stem}\t{o}\n')

        log(f'Ready to annotate {len(CONTIG_FILES)} genomes in dir "{BASE_DIR}" with '
                  f'{CPUS_PER_GENOME} threads per genome and tasks {TASKS}.')

def spawn_file(program_name, genome_id, base_dir = None) -> Path:
    """computes a Path genome_id.program_name or, if multimode==True, program_name/genome_id"""
    target_dir = Path(base_dir) if base_dir else TEMP_DIR
    if MULTI_MODE:
        dir = target_dir / program_name
        dir.mkdir(exist_ok=True)
        if dir.is_file():
            if FORCE:
                dir.unlink()
                dir.mkdir(exist_ok=True)
            else:
                raise Exception("Use force to overwrite existing results")
        return dir / genome_id
    else:
        file = target_dir / f'{genome_id}.{program_name}'
        if file.exists() and file.is_dir():
            if FORCE:
                shutil.rmtree(file)
        return file


def log(log_message, values=(), topic=''):
    if not topic or topic in LOG_TOPICS:
        if len(values):
            final_msg = f'{format_runtime()} {log_message.format(*values)}'
        else:
            final_msg = f'{format_runtime()} {log_message}'
        print(final_msg)
        try:
            with open(LOG_FILE, 'a') as log_handle:
                log_handle.write(final_msg)
                log_handle.write('\n')
        except FileNotFoundError:
            pass


def format_runtime():
    runtime = time.monotonic() - START_TIME
    return f'[{int(runtime / 3600):02d}h:{int((runtime % 3600) / 60):02d}m:{int(runtime % 60):02d}s]'


def run_external(exec, stdin=None, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, log_cmd=True):
    if log_cmd:
        log(exec)
    result = subprocess.run(exec.split(), stdout=stdout, stdin=stdin, stderr=stderr)
    if result.returncode != 0:
        raise Exception(f'WARNING: "{exec}" exited with non-zero status')


def download(url: str, file: Path):
    log(f'Started downloading {url} to {file}...')
    with httpx.stream('GET', url, timeout=6.1, follow_redirects=True) as data_stream, open(file, 'wb') as handle:
        for data in data_stream.iter_bytes():
            handle.write(data)


def sorted_annotators():
    return (registry.ANNOTATOR_REGISTRY[a] for a in sorted(list(registry.ANNOTATOR_REGISTRY.keys())))


def parse_metaerg_progress(genome_name):
    progress_file = spawn_file('metaerg_progress', genome_name)
    progress = ''
    if progress_file.exists():
        with open(progress_file) as progress_reader:
            for line in progress_reader:
                progress += line
    return progress


def write_metaerg_progress(genome_name, new_progress):
    progress_file = spawn_file('metaerg_progress', genome_name)
    with open(progress_file, 'w') as writer:
        writer.write(new_progress)


def register_annotator(define_annotator):
    param = define_annotator()

    def annotator(genome_name, contig_dict, db_connection) -> int:
        """Runs programs and reads results."""
        # (1) Read metaerg progress file, update progress and log the start of the analysis
        if not param['annotator_key'] in ACTIVE_ANNOTATORS:
            return 0
        if FORCE:
            write_metaerg_progress(genome_name, '')
        current_progress = parse_metaerg_progress(genome_name)
        if not '{}=complete'.format(param['annotator_key']) in current_progress:
            current_progress += '{}=started\n'.format(param['annotator_key'])
            write_metaerg_progress(genome_name, current_progress)
        log('({}) Starting {} ...', (genome_name, param['purpose']))

        # (2) Make sure required databases are available
        for d in param.get('databases', []):
            d = DATABASE_DIR / d
            if not d.exists() or not d.stat().st_size:
                log('({}) Unable to run {}, or parse results, database "{}" missing', (genome_name,
                                                                                       param['purpose'], d))
                return 0
        # (3) Then, if force or the results files are not yet there, run the programs:
        result_files = [spawn_file(f, genome_name) for f in param.get('result_files', [])]
        analysis_already_completed = True
        for f in result_files:
            if not f.exists() or not f.stat().st_size:
                analysis_already_completed = False
                break
        # (4) check if we tried to run this analysis before without completing...
        if '{}=started'.format(param['annotator_key']) in current_progress:
            analysis_already_completed = False
        elif '{}=complete'.format(param['annotator_key']) in current_progress:
            analysis_already_completed = True

        if not READ_ONLY and (FORCE or not analysis_already_completed):
            # (5) make sure that the helper programs are available:
            all_programs_in_path = True
            p = 'X'
            for p in param.get('programs', []):
                program_path = shutil.which(p, mode=os.X_OK)
                if not program_path:
                    all_programs_in_path = False
            if all_programs_in_path:
                try:
                    param['run'](genome_name, contig_dict, db_connection, result_files)
                except Exception as e:
                    log('({}) Error while running {}: {}', (genome_name, param['purpose'],
                                                            "".join(traceback.format_exception(e))))
                    return 0
            else:
                log('({}) Unable to run {}, helper program "{}" not in path', (genome_name, param['purpose'], p))
                return 0
        elif len(param.get('programs', [])):  # before logging this, do check if any result files will be parsed
            log('({}) Reusing existing results in {}.'. format(genome_name,
                                                              ', '.join(str(file) for file in result_files)))
        # (6) Check if all results files are there, otherwise return:
        results_complete = True
        for f in result_files:
            if not f.exists() or not f.stat().st_size:
                log('({}) Missing expected result file {}; this could mean no results were found or that a '
                    'helper program failed to run.', (genome_name, f))
                results_complete = False
        # (7) Report success, update progress
        current_progress = current_progress.replace('{}=started'.format(param['annotator_key']),
                                                    '{}=complete'.format(param['annotator_key']))
        write_metaerg_progress(genome_name, current_progress)
        positive_count = 0
        if results_complete:
            positive_count = param['read'](genome_name, contig_dict, db_connection, result_files)
        log('({}) {} complete. Found {}.', (genome_name, param['purpose'], positive_count))
        return 0

    registry.ANNOTATOR_REGISTRY[param['pipeline_position']] = annotator
    return annotator


def register_html_writer(writer):
    registry.HTML_WRITER_REGISTRY.append(writer)
    return writer


def register_database_installer(database_installer):
    registry.DATABASE_INSTALLER_REGISTRY.append(database_installer)
    # print(len(registry.DATABASE_INSTALLER_REGISTRY))
    return database_installer



import gzip


class DBentry:
    FIELDS = ('domain', 'descr', 'taxon', 'accession', 'gene', 'length', 'pos', 'min_score', 'min_t_score')

    def __init__(self, *, domain: str, descr: str, taxon: str = '', accession: str = '', gene: str = '',
                 length: int = 0, pos: int = 0, min_score: int = 0, min_t_score: int = 0):
        self.domain = domain
        self.descr = descr
        self.taxon = taxon
        self.accession = accession
        self.gene = gene
        self.length = length
        self.pos = pos
        self.min_score = min_score
        self.min_t_score = min_t_score

    def __iter__(self):
        return ((k, v) for k, v in zip(DBentry.FIELDS, (self.domain, self.descr, self.taxon, self.accession,
                                                        self.gene, self.length, self.pos, self.min_score,
                                                        self.min_t_score)))

    def __repr__(self):
        return '{}({})'.format(type(self).__name__, ', '.join(f'{k}={v!r}' for (k, v) in self))

    def __len__(self):
        return self.length


def taxon_at_genus(taxon) -> str:
    if type(taxon) == str:
        for t in reversed(taxon.split("; ")):
            if " " not in t:
                return t
    return ''


class BlastHit:
    FIELDS = ('query', 'hit', 'percent_id', 'aligned_length', 'mismatches', 'gaps', 'query_start',
              'query_end', 'hit_start', 'hit_end', 'evalue', 'score')

    def __init__(self, query: str, hit: DBentry, percent_id: float, aligned_length: int,
                 query_start: int, query_end: int, hit_start: int, hit_end: int, evalue: float, score: float,
                 mismatches: int = 0, gaps: int = 0):
        self.query = query
        self.hit = hit
        self.percent_id = percent_id
        self.aligned_length = aligned_length
        self.mismatches = mismatches
        self.gaps = gaps
        self.query_start = query_start
        self.query_end = query_end
        self.hit_start = hit_start
        self.hit_end = hit_end
        self.evalue = evalue
        self.score = score

    def __iter__(self):
        return ((k, v) for k, v in zip(BlastHit.FIELDS, (self.query, self.hit, self.percent_id, self.aligned_length,
                                                         self.mismatches, self.gaps, self.query_start, self.query_end,
                                                         self.hit_start, self.hit_end, self.evalue, self.score)))

    def __repr__(self):
        return '{}({})'.format(type(self).__name__, ', '.join(f'{k}={v!r}' for (k, v) in self))

    def __len__(self):
        return max(self.aligned_length, 1)


class BlastResult:
    def __init__(self, hits: tuple[BlastHit]):
        self.hits = hits
        if not len(hits):
            raise Exception('Attempt to create empty blast result.')

    def __iter__(self):
        return self.hits.__iter__()

    def __len__(self):
        return len(self.hits)

    def __repr__(self):
        return '{}(({},))'.format(type(self).__name__, ',\n'.join(f'{h!r}' for h in self))

    def query(self):
        return self.hits[0].query

    def percent_aligned(self) -> float:
        return 100 * min(1.0, len(self.hits[0]) / len(self.hits[0].hit))

    def percent_recall(self) -> float:
        return 100 * sum((1 for h in self.hits[1:] if h.hit.descr == self.hits[0].hit.descr)) / len(self)

    def summary(self) -> str:
        identical_function_count = sum((1 for h in self.hits[1:] if h.hit.descr == self.hits[0].hit.descr))
        return '[{}/{}] aa@{}% [{}/{}] {}'.format(len(self.hits[0]),
                                                  len(self.hits[0].hit),
                                                  self.hits[0].percent_id,
                                                  identical_function_count,
                                                  len(self),
                                                  self.hits[0].hit.descr)


class TabularBlastParser:
    def __init__(self, filename, mode, retrieve_db_entry):
        self.filename = filename
        self.mode = mode
        self.handle = None
        self.retrieve_db_entry = retrieve_db_entry

    def __enter__(self):
        if str(self.filename).endswith('.gz'):
            self.handle = gzip.open(self.filename, 'rt')
        else:
            self.handle = open(self.filename)
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.handle.close()

    def __iter__(self):
        all_hits: list[BlastHit] = []
        while next_hit := self.load_next_hit_from_file():
            if not len(all_hits) or all_hits[-1].query == next_hit.query:
                all_hits.append(next_hit)
            else:
                yield BlastResult(tuple(all_hits))
                all_hits = [next_hit]
        if len(all_hits):
            yield BlastResult(tuple(all_hits))

    def load_next_hit_from_file(self) -> BlastHit | None:
        while line := self.handle.readline():
            words = line.strip().split()  # \t works for blast, but hmmer and cmscan use spaces
            match words:
                case [word, *_] if word.startswith('#'):
                    continue
                case [query, hit, percent_id, aligned_length, mismatches, gaps, query_start, query_end, hit_start,
                      hit_end, evalue, score] if 'BLAST' == self.mode:
                    hit_db_entry = self.retrieve_db_entry(hit)
                    b = BlastHit(query, hit_db_entry, float(percent_id), int(aligned_length), int(query_start),
                                 int(query_end), int(hit_start), int(hit_end), float(evalue), float(score),
                                 int(mismatches), int(gaps))
                    return(b)
                case [hit, _, query, _, evalue, score, _, _, _, _, _, _, _, _, _, _, _, _, *_] if 'HMMSCAN' == self.mode:
                    hit_db_entry = self.retrieve_db_entry(hit)
                    return BlastHit(query, hit_db_entry, 0, hit_db_entry.length, 0, 0, 0, 0, float(evalue), float(score), 0, 0)
                case [hit, _, _, query, _, _, evalue, score, _, _, _, _, _, _, _, hit_start, hit_end,
                      query_start, query_end, *_] if 'HMMSCAN_DOM_TABLE' == self.mode:
                    hit_db_entry = self.retrieve_db_entry(hit)
                    hit_start = int(hit_start)
                    hit_end = int(hit_end)
                    return BlastHit(query, hit_db_entry, 0, hit_end - hit_start, int(query_start), int(query_end),
                                    hit_start, hit_end, float(evalue), float(score))

                case [query, _, hit, _, evalue, score, _, _, _, _, _, _, _, _, _, _, _, _, *_] if 'HMMSEARCH' == self.mode:
                    hit_db_entry = self.retrieve_db_entry(hit)
                    return BlastHit(query, hit_db_entry, 0, hit_db_entry.length, 0, 0, 0, 0, float(evalue), float(score), 0, 0)
                case [*_]:
                    continue
        return None



import shutil
import os
from pathlib import Path

from metaerg import context
from metaerg.datatypes import gbk
from metaerg.datatypes import sqlite


def _run_programs(genome_name, contig_dict, db_connection, result_files):
    """Should execute the helper programs to complete the analysis"""
    gbk_file = context.spawn_file('gbk', genome_name)
    if context.MULTI_MODE:
        gbk_file = Path(gbk_file.parent, gbk_file.name + '.gbk')
    with open(gbk_file, 'w') as handle:
        gbk.gbk_write_genome(handle, contig_dict, db_connection)
    context.run_external(f'antismash --genefinding-tool none --output-dir {result_files[0]} {gbk_file}')


def _read_results(genome_name, contig_dict, db_connection, result_files) -> int:
    """Should parse the result files and return the # of positives."""
    antismash_hit_count = 0
    for f in sorted(result_files[0].glob('*region*.gbk')):
        with gbk.GbkFeatureParser(f) as reader:
            antismash_region_name = ''
            antismash_region_number = 0
            for f_as_dict in reader:
                if 'region' == f_as_dict['type']:
                    antismash_region_name = '{} {}'.format(f_as_dict['product'], f_as_dict['rules'])
                    antismash_region_number = int(f_as_dict['region_number'])
                elif 'CDS' == f_as_dict['type']:
                    antismash_gene_function = f_as_dict.get('gene_functions', '')
                    antismash_gene_category = f_as_dict.get('gene_kind', '')
                    if antismash_region_name:
                        feature = sqlite.read_feature_by_id(db_connection, f_as_dict['locus_tag'])
                        if not feature:
                            # antismash sometimes predicts a new ORF, for example for a lassopeptide
                            # we do not add this to the feature table
                            continue
                        feature.antismash =  ' '.join((f'(region {antismash_region_number})', antismash_region_name,
                                                       antismash_gene_function, antismash_gene_category))
                        feature.subsystems = (feature.subsystems + ' [secondary-metabolites]').strip()
                        sqlite.update_feature_in_db(db_connection, feature)
                        antismash_hit_count += 1
    return antismash_hit_count


@context.register_annotator
def run_and_read_antismash():
    return ({'pipeline_position': 91,
             'annotator_key': 'antismash',
             'purpose': 'prediction of secondary metabolite genes with antismash',
             'programs': ('antismash',),
             'result_files': ('antismash',),
             'run': _run_programs,
             'read': _read_results})


@context.register_html_writer
def write_html(genome_name, db_connection, genome_properties:dict, dir):
    """need to copy the antismash result dir to the metaerg html dir."""
    dir.mkdir(exist_ok=True, parents=True)
    antismash_result_dir = context.spawn_file('antismash', genome_name)
    antismash_html_parent = Path(dir, genome_name, 'antismash')
    if antismash_html_parent.exists():
        shutil.rmtree(antismash_html_parent)
    if antismash_result_dir.exists():
        shutil.copytree(antismash_result_dir, antismash_html_parent)


@context.register_database_installer
def format_antismash_databases():
    if 'A' not in context.TASKS:
        return
    antismash_database_path = context.DATABASE_DIR / 'antismash'
    context.log(f'Installing antismash database at {antismash_database_path}')
    antismash_database_path.mkdir(parents=True, exist_ok=True)
    os.system(f'download-antismash-databases --database-dir {antismash_database_path}')



from metaerg import context
from metaerg.datatypes import fasta, gff
from metaerg.datatypes import gff
from metaerg.datatypes import sqlite


def _run_programs(genome_name, contig_dict, db_connection, result_files):
    fasta_file = context.spawn_file('masked', genome_name)
    fasta.write_contigs_to_fasta(contig_dict, fasta_file, db_connection, genome_name,
                                 mask_targets=fasta.ALL_MASK_TARGETS)
    ltr_index_file = context.spawn_file('ltr_index', genome_name)

    context.run_external(f'gt suffixerator -db {fasta_file} -indexname {ltr_index_file} -tis -suf -lcp -des -ssp -sds -dna')
    context.run_external(f'gt ltrharvest -index {ltr_index_file} -gff3 {result_files[0]} -seqids')
    # remove index files
    for file in ltr_index_file.parent.glob(f'{genome_name}.ltr_index*'):
        file.unlink()


def _read_results(genome_name, contig_dict, db_connection, result_files) -> int:
    count = 0
    with gff.GffParser(result_files[0], contig_dict, inference='ltr_harvest',
                       target_feature_type_dict={'repeat_region': 'retrotransposon'}) as parser:
        for feature in parser:
            sqlite.add_new_feature_to_db(db_connection, feature)
            count += 1
    return count


@context.register_annotator
def run_and_read_ltr_harvest():
    return ({'pipeline_position': 31,
             'annotator_key': 'ltr_harvest',
             'purpose': 'retrotransposon prediction with ltrharvest',
             'programs': ('gt',),
             'result_files': ('ltr_harvest',),
             'run': _run_programs,
             'read': _read_results})



import gzip
import re
import numpy as np
from pathlib import Path
from metaerg import context
from metaerg.datatypes import sqlite

NON_IUPAC_RE_NT = re.compile(r'[^ACTGN]')
NON_IUPAC_RE_AA = re.compile(r'[^RHKDESTNQCUGPAVILMFYW]')
ALL_MASK_TARGETS = set('CDS rRNA tRNA tmRNA ncRNA repeat crispr_repeat retrotransposon'.split())
COMPLEMENT = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A',}


def reverse_complement(seq:str) -> str:
    return ''.join(COMPLEMENT.get(b, 'N') for b in reversed(seq))


def _parse_fasta_header(line:str) -> dict:
    si = line.find(' ')
    if si > 0:
        return {'id': line[1:si], 'seq': '', 'descr': line[si + 1:].strip()}
    else:
        return {'id': line[1:].strip(), 'seq': '', 'descr': ''}


class FastaParser:
    def __init__(self, path, cleanup_seq=True):
        self.path = path
        self.handle = None
        self.cleanup_seq = cleanup_seq
        self.alphabet = None
        self.unknown_char = ''

    def __enter__(self):
        if str(self.path).endswith('.gz'):
            self.handle = gzip.open(self.path, 'rt')
        else:
            self.handle = open(self.path)
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.handle.close()

    def __iter__(self):
        seq_rec = None
        seq = []
        while line := self.handle.readline():
            line = line.strip()
            if line.startswith('#'):
                continue
            if line.startswith('>'):
                if len(seq) and seq_rec is not None:
                    seq_rec['seq'] = self._cleanup(''.join(seq))
                    seq = []
                    yield seq_rec
                seq_rec = _parse_fasta_header(line)
            elif seq_rec is not None:
                seq.append(line)
                if len(line) > 10 and not self.alphabet and self.cleanup_seq:
                    seq_nt_errors = sum(1 for match in NON_IUPAC_RE_NT.finditer(line))
                    seq_aa_errors = sum(1 for match in NON_IUPAC_RE_AA.finditer(line))
                    if seq_nt_errors <= seq_aa_errors:
                        self.alphabet = NON_IUPAC_RE_NT
                        self.unknown_char = 'N'
                    else:
                        self.alphabet = NON_IUPAC_RE_AA
                        self.unknown_char = 'X'
        if seq_rec is not None:
            seq_rec['seq'] = self._cleanup(''.join(seq))
            yield seq_rec

    def _cleanup(self, seq) -> str:
        if self.cleanup_seq:
            seq = seq.upper()
            seq = ''.join(seq.split())
            if self.alphabet:
                seq = self.alphabet.sub(self.unknown_char, seq)
        return seq


class Masker:
    def __init__(self, db_connection, targets=None):
        self.db_connection = db_connection
        self.targets = targets
        self.nt_total = 0
        self.nt_masked = 0

    def mask(self, seq_record):
        seq = seq_record['seq']
        if self.targets:
            for feature in sqlite.read_all_features(self.db_connection,  contig=seq_record['id'], type=self.targets):
                seq = seq[:feature.start] + 'N' * feature.length_nt() + seq[feature.end:]
                self.nt_masked += feature.length_nt()

        self.nt_total += len(seq)

        masked_record = seq_record.copy()
        masked_record['seq'] = seq
        return masked_record

    def stats(self):
        return f'Masked {self.nt_masked} / {self.nt_total} nt ({self.nt_masked / max(self.nt_total, 1):.1%}).'


def write_fasta(handle, fasta, line_length=80):
    if not fasta:
        raise Exception('Attempt to write zero-length sequence to fasta.')
    handle.write(f'>{fasta["id"]}')
    try:
        if fasta["descr"] and fasta["descr"] is not np.nan:
            handle.write(f' {fasta["descr"]}')
    except KeyError:
        pass
    try:
        if fasta["subsystems"]:
            handle.write(f' ({fasta["subsystems"]})')
    except KeyError:
        pass
    try:
        if fasta["taxon"]:
            handle.write(f' ({fasta["taxon"]})')
    except KeyError:
        pass
    handle.write('\n')
    for i in range(0, len(fasta['seq']), line_length):
        handle.write(fasta['seq'][i:i+line_length])
        handle.write('\n')


def write_features_to_fasta(db_connection, seq_type: str, base_file: Path, split=1, targets = None):
    number_of_records = sum(1 for f in sqlite.read_all_features(db_connection, type=targets))
    split = min(split, number_of_records)
    records_per_file = number_of_records / split if split else number_of_records
    if split > 1:
        paths = [Path(base_file.parent, f'{base_file.name}.{i}') for i in range(split)]
    else:
        paths = base_file,
    filehandles = [open(p, 'w') for p in paths]
    records_written = 0
    for feature in sqlite.read_all_features(db_connection, type=targets):
        fasta_rec = {'id': feature.id, 'descr': feature.descr}
        if 'aa' == seq_type:
            fasta_rec['seq'] = feature.aa_seq
        elif 'nt' == seq_type:
            fasta_rec['seq'] = feature.nt_seq
        write_fasta(filehandles[int(records_written / records_per_file)], fasta_rec)
        records_written += 1
    if not number_of_records:
        context.log(f'WARNING: No [{", ".join(targets)}] features in genome, '
                    f'fasta file(s) {", ".join([str(p.name) for p in paths])} are empty!')
    return paths


def write_contigs_to_fasta(contig_dict: dict, base_file: Path, db_connection, genome_name='',
                           split=1, mask_targets=None):
    """writes contigs to fasta file(s), optionally masking features with N"""
    number_of_records = len(contig_dict)
    split = int(min(split, number_of_records))
    records_per_file = number_of_records / split
    if split > 1:
        paths = [Path(base_file.parent, f'{base_file.name}.{i}') for i in range(split)]
    else:
        paths = base_file,
    filehandles = [open(p, 'w') for p in paths]
    records_written = 0
    masker = Masker(db_connection, targets=mask_targets)
    for contig in contig_dict.values():
        write_fasta(filehandles[int(records_written / records_per_file)], masker.mask(contig))
        records_written += 1
    for f in filehandles:
        f.close()
    if mask_targets:
        context.log(f'({genome_name}) {masker.stats()}')
    return paths



import shutil
from pathlib import Path
from concurrent.futures import ProcessPoolExecutor

from metaerg import context
from metaerg.datatypes import fasta
from metaerg.datatypes import sqlite


def _run_programs(genome_name, contig_dict, db_connection, result_files):
    cds_aa_file = context.spawn_file('cds.faa', genome_name)
    if context.CPUS_PER_GENOME > 1:
        split_fasta_files = fasta.write_features_to_fasta(db_connection, 'aa', cds_aa_file, context.CPUS_PER_GENOME,
                                                          targets=('CDS',))
        split_signalp_files = [result_files[0].parent / f'{result_files[0].name}.{i}'
                               for i in range(len(split_fasta_files))]
        with ProcessPoolExecutor(max_workers=context.CPUS_PER_GENOME) as executor:
            for split_input, split_output in zip(split_fasta_files, split_signalp_files):
                executor.submit(context.run_external, f'signalp6 --fastafile {split_input} --output_dir '
                                                    f'{split_output} --format none --organism other')

        result_files[0].mkdir(exist_ok=True)
        with open(result_files[0] / 'prediction_results.txt', 'wb') as output:
            for split_cds_aa_file, split_signalp_dir in zip(split_fasta_files, split_signalp_files):
                signalp_result_file = Path(split_signalp_dir, 'prediction_results.txt')
                if signalp_result_file.exists():
                    with open(signalp_result_file, 'rb') as input:
                        shutil.copyfileobj(input, output)
                else:
                    context.log(f'({genome_name}) WARNING - missing part of signalp output!')
                if split_signalp_dir.exists():
                    shutil.rmtree(split_signalp_dir)
                split_cds_aa_file.unlink(missing_ok=True)
    else:
        context.run_external(f'signalp6 --fastafile {cds_aa_file} --output_dir {result_files[0]} --format none --organism other')


def _read_results(genome_name, contig_dict, db_connection, result_files) -> int:
    count = 0
    signalp_result_file = result_files[0] / 'prediction_results.txt'
    if signalp_result_file.exists():
        with open(signalp_result_file) as signalp_handle:
            for line in signalp_handle:
                if line.startswith("#"):
                    continue
                words = line.split("\t")
                if "OTHER" == words[1]:
                    continue
                feature_id = words[0].split()[0]
                feature = sqlite.read_feature_by_id(db_connection, feature_id)
                if not feature:
                    raise Exception(f'Found results for unknown feature {feature_id}, '
                                    f'may need to rerun metaerg with --force')
                feature.signal_peptide = words[1]
                sqlite.update_feature_in_db(db_connection, feature)
                count += 1
    return count


@context.register_annotator
def run_and_read_signalp():
    return ({'pipeline_position': 121,
             'annotator_key': 'signalp',
             'purpose': 'signal peptide prediction with signalp',
             'programs': ('signalp6',),
             'result_files': ('signalp',),
             'run': _run_programs,
             'read': _read_results})



import os
import shutil
from pathlib import Path
from concurrent.futures import ProcessPoolExecutor

from metaerg import context
from metaerg import functional_gene_configuration
from metaerg.datatypes import fasta
from metaerg.datatypes import sqlite
from metaerg.datatypes.blast import DBentry, TabularBlastParser


def _run_programs(genome_name, contig_dict, db_connection, result_files):
    cds_aa_file = context.spawn_file('cds.faa', genome_name)
    cdd_database = Path(context.DATABASE_DIR, 'cdd', 'Cdd')
    if context.CPUS_PER_GENOME > 1:
        split_fasta_files = fasta.write_features_to_fasta(db_connection, 'aa', cds_aa_file, targets=('CDS',),
                                                               split=context.CPUS_PER_GENOME)
        split_cdd_files = [Path(result_files[0].parent, f'{result_files[0].name}.{i}')
                           for i in range(len(split_fasta_files))]
        with ProcessPoolExecutor(max_workers=context.CPUS_PER_GENOME) as executor:
            for split_input, split_output in zip(split_fasta_files, split_cdd_files):
                executor.submit(context.run_external, f'rpsblast -db {cdd_database} -query {split_input} '
                                                      f'-out {split_output} -outfmt 6 -evalue 1e-7')

        with open(result_files[0], 'wb') as output:
            for split_input_file, split_output_file in zip(split_fasta_files, split_cdd_files):
                with open(split_output_file, 'rb') as input:
                    shutil.copyfileobj(input, output)
                split_input_file.unlink()
                split_output_file.unlink()
    else:
        context.run_external(f'rpsblast -db {cdd_database} -query {cds_aa_file} -out {result_files[0]} -outfmt 6 -evalue 1e-7')


def _read_results(genome_name, contig_dict, db_connection, result_files) -> int:
    # load cdd index
    cdd = {}
    cdd_index = Path(context.DATABASE_DIR, 'cdd', 'cddid.tbl')
    with open(cdd_index) as db_handle:
        for line in db_handle:
            words = line.strip().split("\t")
            cdd[int(words[0])] = DBentry(domain='cdd', accession=words[1], gene=words[2], descr=words[3],
                                         length=int(words[4]))
    context.log(f'({genome_name}) Parsed {len(cdd)} entries from conserved domain database.')
    # parse cdd results
    cdd_result_count = 0
    def get_cdd_db_entry(id: str) -> DBentry:
        return cdd[int(id[4:])]

    with TabularBlastParser(result_files[0], 'BLAST', get_cdd_db_entry) as handle:
        for cdd_result in handle:
            feature = sqlite.read_feature_by_id(db_connection, cdd_result.query())
            if not feature:
                raise Exception(f'Found results for unknown feature {cdd_result.query()}, '
                                f'may need to rerun metaerg with --force')
            feature.cdd = cdd_result
            cdd_result_count += 1
            if new_subsystem := functional_gene_configuration.match(cdd_result):
                feature.subsystems = functional_gene_configuration.cleanup_subsystem_str(
                    feature.subsystems.strip() + ' ' + new_subsystem)
            top_entry: DBentry = cdd_result.hits[0].hit
            feature.descr = f'{top_entry.accession}|{top_entry.gene} {top_entry.descr}'
            if len(feature.descr) > 35:
                feature.descr = feature.descr[:35] + '...'
            sqlite.update_feature_in_db(db_connection, feature)
    return cdd_result_count


@context.register_annotator
def run_and_read_cdd():
    return ({'pipeline_position': 71,
             'annotator_key': 'cdd',
             'purpose': 'function prediction using RPSBlast and the conserved domain database',
             'programs': ('rpsblast',),
             'databases': (Path('cdd', 'cddid.tbl'), Path('cdd', 'Cdd.pal')),
             'result_files': ('cdd',),
             'run': _run_programs,
             'read': _read_results})


@context.register_database_installer
def install_cdd_database():
    if 'C' not in context.TASKS:
       return
    cdd_dir = context.DATABASE_DIR / 'cdd'
    context.log(f'Installing the conserved domain database to {cdd_dir}...')
    cdd_dir.mkdir(exist_ok=True, parents=True)
    cdd_index = cdd_dir / 'cddid.tbl'
    if context.FORCE or (not cdd_index.exists() and not (cdd_dir / 'cddid.tbl.gz').exists()):
        context.run_external(f'wget -P {cdd_dir} https://ftp.ncbi.nih.gov/pub/mmdb/cdd/cddid.tbl.gz')
    if context.FORCE or not cdd_index.exists():
        context.run_external(f'gunzip {cdd_index}.gz')

    temp_cdd_dir = context.DATABASE_DIR / 'cdd-temp'
    temp_cdd_dir.mkdir(exist_ok=True, parents=True)
    if context.FORCE or (not (temp_cdd_dir / 'cdd.tar').exists() and not (temp_cdd_dir / 'cdd.tar.gz').exists()):
        context.run_external(f'wget -P {temp_cdd_dir} https://ftp.ncbi.nih.gov/pub/mmdb/cdd/cdd.tar.gz')
    if context.FORCE or not (temp_cdd_dir / 'Tigr.pn').exists():
        context.run_external(f'tar -xf {temp_cdd_dir / "cdd.tar.gz"} -C {temp_cdd_dir}')
    current_dir = os.getcwd()
    os.chdir(temp_cdd_dir)
    if context.FORCE or not (cdd_dir / 'Cdd.pal').exists():
        context.run_external(f'makeprofiledb -title CDD.v.3.12 -in {temp_cdd_dir / "Cdd.pn"} -out '
                             f'{cdd_dir / "Cdd"} -threshold 9.82 -scale 100.0 -dbtype rps '
                             f'-index true')
    os.chdir(current_dir)



import shutil
from pathlib import Path

from metaerg import context
from metaerg.datatypes import fasta
from metaerg.datatypes import sqlite

def _run_programs(genome_name, contig_dict, db_connection, result_files):
    fasta_file = context.spawn_file('masked', genome_name)
    fasta.write_contigs_to_fasta(contig_dict, fasta_file, db_connection, genome_name,
                                 mask_targets=fasta.ALL_MASK_TARGETS)
    lmer_table_file = context.spawn_file('lmer-table', genome_name)
    repeatscout_file_raw = context.spawn_file('repeatscout-raw', genome_name)
    repeatscout_file_filtered = context.spawn_file('repeatscout-filtered', genome_name)
    repeatscout_file_filtered2 = context.spawn_file('repeatscout-filtered2', genome_name)

    context.run_external(f'build_lmer_table -sequence {fasta_file} -freq {lmer_table_file}')
    context.run_external(f'RepeatScout -sequence {fasta_file} -output {repeatscout_file_raw} -freq {lmer_table_file}')
    with open(repeatscout_file_filtered, 'w') as output, open(repeatscout_file_raw) as input:
        context.run_external('filter-stage-1.prl', stdin=input, stdout=output)

    repeatmasker_output_file = fasta_file.parent / f'{fasta_file.name}.out'  # nothing we can do about that

    if not repeatscout_file_filtered.stat().st_size:
        context.log(f'({genome_name}) No repeats remaining after filter stage 1.')
        finalize(False, genome_name, fasta_file, repeatmasker_output_file, result_files[0])
        return

    context.run_external(f'RepeatMasker -pa {context.CPUS_PER_GENOME} -lib {repeatscout_file_filtered} '
                         f'-dir {fasta_file.parent} {fasta_file}')
    with open(repeatscout_file_filtered2, 'w') as output, open(repeatscout_file_filtered) as input:
        context.run_external(f'filter-stage-2.prl --cat={repeatmasker_output_file} --thresh=10', stdin=input, stdout=output)

    if not repeatscout_file_filtered2.stat().st_size:
        context.log(f'({genome_name}) No repeats remaining after filter stage 2.')
        finalize(False, genome_name, fasta_file, repeatmasker_output_file, result_files[0])
        return

    context.run_external(f'RepeatMasker -pa {context.CPUS_PER_GENOME} -lib {repeatscout_file_filtered2} '
                         f'-dir {fasta_file.parent} {fasta_file}')
    finalize(True, genome_name, fasta_file, repeatmasker_output_file, result_files[0])


def finalize(success: bool, genome_name: str, fasta_input_file: Path, repeatmasker_output_file: Path,
             final_repeatmasker_output_file: Path):
    if not success:
        with open(repeatmasker_output_file, 'w') as handle:
            handle.write('#No repeats detected by repeatmasker')
    shutil.move(repeatmasker_output_file, final_repeatmasker_output_file)
    for file in fasta_input_file.parent.glob(f'{fasta_input_file.name}.*'):
        if file.is_dir():
            shutil.rmtree(file)
        else:
            file.unlink()


def words2feature(words: list[str], contig, genome_name:str) -> sqlite.Feature:
    start = int(words[5]) - 1
    end = int(words[6])
    strand = -1 if 'C' == words[8] else 1
    seq = contig['seq'][start:end]
    if strand < 0:
        seq = fasta.reverse_complement(seq)
    return sqlite.Feature(genome = genome_name,
                          contig = contig['id'],
                          start = start,
                          end = end,
                          strand = strand,
                          type = 'repeat',
                          inference = 'repeatmasker',
                          nt_seq = seq)


def _read_results(genome_name, contig_dict, db_connection, result_files) -> int:
    """(1) simple repeats, these are consecutive
       (2) unspecified repeats, these occur scattered and are identified by an id in words[9]. We only
           add those when they occur 10 or more times."""
    count = 0
    repeat_hash = dict()
    with open(result_files[0]) as repeatmasker_handle:
        for line in repeatmasker_handle:
            words = line.split()
            if len(words) < 11 or words[0] in ('SW', 'score'):
                continue
            try:
                contig = contig_dict[words[4]]
            except KeyError:
                context.log(f'({genome_name}) Warning: Unknown contig id "{words[4]}"')
                continue
            if 'Simple_repeat' == words[10]:
                feature = words2feature(words, contig, genome_name)
                feature.notes = f'repeat {words[9]}'
                sqlite.add_new_feature_to_db(db_connection, feature)
                count += 1
            else:
                repeat_list = repeat_hash.setdefault(words[9], list())
                repeat_list.append(words2feature(words, contig, genome_name))
    for repeat_list in repeat_hash.values():
        if len(repeat_list) >= 10:
            for feature in repeat_list:
                feature.notes = f' (occurs {len(repeat_list)}x)'
                sqlite.add_new_feature_to_db(db_connection, feature)
                count += 1
    return count


@context.register_annotator
def run_and_read_repeatmasker():
    return ({'pipeline_position': 51,
             'annotator_key': 'repeat_masker',
             'purpose': 'repeat prediction with repeatmasker',
             'programs': ('build_lmer_table', 'RepeatScout', 'filter-stage-1.prl', 'RepeatMasker'),
             'result_files': ('repeatmasker',),
             'run': _run_programs,
             'read': _read_results})



import argparse
import tarfile
from pathlib import Path
from concurrent import futures
from hashlib import md5
from collections import Counter

import pandas as pd

import metaerg.run_and_read.diamond_and_blastn
import metaerg.run_and_read.functional_genes
import metaerg.run_and_read.antismash
from metaerg.datatypes import sqlite
from metaerg import context
from metaerg import registry
from metaerg import functional_gene_configuration
from metaerg.datatypes import fasta, gbk
from metaerg.html import html_all_genomes
from metaerg.run_and_read import tmhmm
from metaerg.installation import install_all_helper_programs
from metaerg.calculations.codon_usage_bias import compute_codon_bias_estimate_doubling_time
from metaerg.run_and_read import *
from metaerg.html import *

VERSION = "2.3.19"


def parse_arguments():
    parser = argparse.ArgumentParser(description='metaerg.py. (C) Marc Strous, Xiaoli Dong 2019, 2022')
    parser.add_argument('--contig_file', help='Fasta nucleotide file of the contigs, or dir that '
                                              'contains multiple fasta nucleotide files.')
    parser.add_argument('--database_dir', help='Dir that contains the annotation databases.')
    parser.add_argument('--contig_mode', default=False,  action="store_true",
                        help='Annotate contigs individually instead of assuming they are part of a genome, MAG or bin.')
    parser.add_argument('--rename_contigs', default=False,  action="store_true",
                        help='Renaming contigs can improve visualization and presentation of results.')
    parser.add_argument('--rename_genomes', default=False,  action="store_true",
                        help='Renaming genomes can improve visualization and presentation of results.')
    parser.add_argument('--min_contig_length', default=0,  help='Shorter contigs will be filtered before annotaton.')
    parser.add_argument('--cpus', default=0, help='How many cpus/threads to use (default: all = 0).')
    parser.add_argument('--file_extension', default='.fna', help='When annotating multiple files in a folder, extension'
                                                                 'of the fasta nucleotide files (default: .fna).')
    parser.add_argument('--translation_table', default=11, help='Which translation table to use (default 11).')
    parser.add_argument('--delimiter', default='.', help='Separater character used in feature ids.')
    parser.add_argument('--prefix', default='g', help='Prefix used when renaming genomes (default: g).')
    parser.add_argument('--checkm_dir', default='checkm', help='Dir with the checkm results (default: checkm)')
    parser.add_argument('--gtdbtk_dir', default='gtdbtk', help='Dir with the gtdbtk results (default: gtdbtk).')
    parser.add_argument('--download_database', default=False, action="store_true",
                        help='Download ready-made metaerg database.')
    parser.add_argument('--create_database', default='', help='Create metaerg database from scratch (use "all", '
                                                                 'to create all components of the database.). Use '
                                                                 'any combination of PVEBRCSA to only create specific '
                                                                 'parts of the database (see README)')
    parser.add_argument('--install_deps', default='', help='Dir for installation of all dependencies '
                                                           '(helper programs). Dependencies will be installed here.')
    parser.add_argument('--path_to_signalp', default='', help='Path to signalp-6.0g.fast.tar.gz.')
    parser.add_argument('--path_to_tmhmm', default='', help='Path to tmhmm-2.0c.Linux.tar.gz.')
    parser.add_argument('--force', default=False, action="store_true",
                        help='Use force to overwrite previous result files.')
    parser.add_argument('--read_only', default=False, action='store_true', help="Do not run any helper programs, only "
                                                                                "read results from previous runs.")
    parser.add_argument('--skip_step', default='', help="Skip one or more annotation steps. Steps are: antismash, aragorn, "
                                                        "cdd, cmscan, diamond_and_blastn, hmm, ltr_harvest, minced, prodigal, "
                                                        "signalp, repeat_masker, tmhmm, trf, separated by commas (,)")
    parser.add_argument('--run_step', default='', help="Only (re)run one or more annotation steps. Steps are: antismash, aragorn, "
                                                        "cdd, cmscan, diamond_and_blastn, hmm, ltr_harvest, minced, prodigal, "
                                                        "signalp, repeat_masker, tmhmm, trf, separated by commas (,)")

    return parser.parse_args()


def load_contigs(genome_name, input_fasta_file, delimiter='.', rename_contigs=False, min_contig_length=0):
    if delimiter in genome_name:
        raise Exception(f'Genome id {genome_name} contains "{delimiter}"; change delimiter with '
                                f'--delimiter [new delimiter] or use --rename_genomes')
    names_done = set()
    contigs = list()
    with fasta.FastaParser(input_fasta_file) as fasta_reader:
        for c in fasta_reader:
            if len(c['seq']) < min_contig_length:
                continue
            c['descr'] = ''  # remove descr as it makes parsing results more difficult for some analyses
            contigs.append(c)
            if not rename_contigs:
                if c['id'] in names_done:
                    raise Exception(f'Contig id {c["id"]} not unique. Use --rename_contigs to avoid this problem.')
                names_done.add(c["id"])
    contigs.sort(key=lambda c:len(c['seq']), reverse=True)
    total_length = sum(len(c['seq']) for c in contigs)
    context.log(f'({genome_name}) Loaded {len(contigs)} contigs with total length {total_length:,} from file.')
    if rename_contigs:
        contig_name_mapping_file = context.spawn_file('contig.name.mappings', genome_name)
        context.log(f'({genome_name}) Renaming contigs (see {contig_name_mapping_file})...')
        i = 0
        with open(contig_name_mapping_file, 'w') as mapping_writer:
            for c in contigs:
                new_id = f'{genome_name}.c{i:0>4}'
                mapping_writer.write(f'{c["id"]}\t{new_id}\n')
                c['id'] = new_id
                i += 1
    else:
        for c_id in contigs:
            if delimiter in c_id:
                raise Exception(f'Contig id {c_id} contains "{delimiter}"; change delimiter with '
                                f'--delimiter [new delimiter] or use --rename_contigs')
    contigs = {c['id']: c for c in contigs}
    return contigs


def compute_genome_properties(genome_name: str, input_fasta_file: Path, contig_dict: dict[str, dict], db_connection) -> dict:
    properties = {'genome name':              genome_name,
                  'input fasta file':         input_fasta_file.name,
                  '# total features':         0,
                  '# contigs':                0,
                  '# proteins':               0,
                  '# ribosomal RNA':          0,
                  '# transfer RNA':           0,
                  '# non-coding RNA':         0,
                  '# retrotransposons':       0,
                  '# CRISPR repeats':         0,
                  '# other repeats':          0,
                  '% coding':                 0.0,
                  '% repeats':                0.0,
                  'mean protein length (aa)': 0.0,}
    contigs:list[dict] = list(contig_dict.values())
    contigs.sort(key=lambda c:len(c['seq']), reverse=True)
    properties['size'] = sum(len(c['seq']) for c in contigs)
    properties['# contigs'] = len(contigs)
    properties['% GC'] = sum((c['seq'].count('G') + c['seq'].count('G') for c in contigs)) / \
                               (properties['size'] - sum((c['seq'].count('N') for c in contigs)))
    cum_size = 0
    for c in contigs:
        cum_size += len(c['seq'])
        if cum_size >+ properties['size'] / 2:
            properties['N50'] = len(c['seq'])
            break

    #feature_data = feature_data.assign(length=feature_data['end'] - feature_data['start'])
    #feature_data_cds = feature_data[feature_data['type'] == 'CDS']
    #feature_data_repeats = feature_data[feature_data['type'].isin(('retrotransposon', 'crispr_repeat', 'repeat'))]
    taxon_counts = Counter()
    for feature in sqlite.read_all_features(db_connection):
        properties['# total features'] += 1
        if feature.type == 'CDS':
            properties['# proteins'] += 1
            properties['% coding'] += feature.length_nt()
            properties['mean protein length (aa)'] += feature.length_aa()
            taxon_counts.update((feature.taxon,))
        elif feature.type == 'rRNA':
            properties['# ribosomal RNA'] += 1
            taxon_counts.update((feature.taxon,))
        elif feature.type == 'tRNA':
            properties['# transfer RNA'] += 1
            taxon_counts.update((feature.taxon,))
        elif feature.type == 'ncRNA':
            properties['# non-coding RNA'] += 1
            taxon_counts.update((feature.taxon,))
        elif feature.type == 'retrotransposon':
            properties['# retrotransposons'] += 1
            properties['% repeats'] += feature.length_nt()
        elif feature.type == 'crispr_repeat':
            properties['# CRISPR repeats'] += 1
            properties['% repeats'] += feature.length_nt()
        elif feature.type == 'repeat':
            properties['# other repeats'] += 1
            properties['% repeats'] += feature.length_nt()
    properties['% coding'] /= properties['size']
    properties['mean protein length (aa)'] /= properties['# proteins']
    properties['% repeats'] /= properties['size']
    properties['% CDS classified to taxon'] = 1 - taxon_counts[''] / taxon_counts.total()
    properties['dominant taxon'] = ''
    properties['% of CDS classified to dominant taxon'] = 0.0
    del taxon_counts['']
    for k, v in taxon_counts.most_common(1):
        properties['% of CDS classified to dominant taxon'] = v / taxon_counts.total()
        properties['dominant taxon'] = k
        break
    codon_usage_bias, doubling_time = compute_codon_bias_estimate_doubling_time(db_connection)
    properties['codon usage bias'] = codon_usage_bias
    properties['doubling_time (days)'] = doubling_time
    properties['subsystems'] = functional_gene_configuration.aggregate(db_connection)
    return properties


def annotate_genome(genome_name, input_fasta_file: Path):
    context.log(f'Started annotation of {genome_name}...')
    current_progress = context.parse_metaerg_progress(genome_name)
    if not context.FORCE and 'visualization=complete' in current_progress:
        context.log(f'({genome_name}) already completed!')
        return
    # (1) prepare sqlite3 database
    db_file = context.spawn_file('annotations.sqlite', genome_name, context.BASE_DIR)
    sqlite.create_db(db_file)
    db_connection = sqlite.connect_to_db(db_file)
    # (2) load sequence data
    contig_dict = load_contigs(genome_name, input_fasta_file, delimiter=context.DELIMITER,
                               min_contig_length=context.MIN_CONTIG_LENGTH, rename_contigs=context.RENAME_CONTIGS)
    # (3) now annotate
    for annotator in context.sorted_annotators():
        annotator(genome_name, contig_dict, db_connection)
    # feather_file = context.spawn_file("all_genes.feather", genome_name, context.BASE_DIR)
    # feature_data = pd.read_feather(feather_file)
    # feature_data = feature_data.set_index('id', drop=False)

    # (4) save results
    context.log(f'({genome_name}) Now writing annotations to .fasta, .gbk...')
    faa_file = context.spawn_file("faa", genome_name, context.BASE_DIR)
    rna_file = context.spawn_file("rna.fna", genome_name, context.BASE_DIR)
    gbk_file = context.spawn_file("gbk", genome_name, context.BASE_DIR)
    fna_file = context.spawn_file("fna", genome_name, context.BASE_DIR)
    fasta.write_features_to_fasta(db_connection, 'aa', faa_file, targets=('CDS',))
    fasta.write_features_to_fasta(db_connection, 'nt', rna_file, targets=('rRNA tRNA tmRNA ncRNA retrotransposon'.split()))
    fasta.write_contigs_to_fasta(contig_dict, fna_file, db_connection)
    with open(gbk_file, 'w') as gbk_writer:
        gbk.gbk_write_genome(gbk_writer, contig_dict, db_connection)
    feature_count = sqlite.count_features(db_connection)
    if feature_count < 1e6:
        context.log(f'({genome_name}) Writing {feature_count} annotations to .feather format...')
        feather_file = context.spawn_file("annotations.feather", genome_name, context.BASE_DIR)
        rows = []
        for feature in sqlite.read_all_features(db_connection):
            rows.append({k: str(v) for k, v in feature})
        pd.DataFrame(rows).to_feather(feather_file)
    else:
        context.log(f'({genome_name}) Skipping feather format, too many ({feature_count}) annotations...')
    # (5) visualize
    genome_properties = compute_genome_properties(genome_name, input_fasta_file, contig_dict, db_connection)
    context.log(f'({genome_name}) Now writing final result as .html for visualization...')
    for html_writer in registry.HTML_WRITER_REGISTRY:
        html_writer(genome_name, db_connection, genome_properties, context.HTML_DIR)
    # (5) update progress
    current_progress = context.parse_metaerg_progress(genome_name)
    current_progress += 'visualization=complete\n'
    context.write_metaerg_progress(genome_name, current_progress)
    context.log(f'({genome_name}) Completed html visualization.')
    return genome_properties


def write_functional_genes_to_xls(genome_properties: dict):
    excel_file = context.BASE_DIR / 'functional_genes.xls'
    if not genome_properties:
        db_files = [context.spawn_file('annotations.sqlite', genome_name, context.BASE_DIR)
                         for genome_name in context.GENOME_NAMES]
        fna_files = [context.spawn_file("fna", genome_name, context.BASE_DIR)
                     for genome_name in context.GENOME_NAMES]
        genome_name_mappings = {}
        with open(context.GENOME_NAME_MAPPING_FILE) as name_mappings:
            for line in name_mappings:
                w = line.split()
                genome_name_mappings[w[0]] = Path(w[2])
        genome_properties = {}
        for db_file, contig_file in zip(db_files, fna_files):
            genome_name = contig_file.stem
            contig_dict = load_contigs(genome_name, contig_file, delimiter='xxxx')
            db_connection = sqlite.connect_to_db(db_file)
            genome_properties[genome_name] = compute_genome_properties(genome_name,
                                                                       genome_name_mappings.get(genome_name, 'unknown'),
                                                                       contig_dict, db_connection)

    all_genome_feature_data = None
    for genome_name, genome_property_hash in genome_properties.items():
        subsystems_df = genome_property_hash['subsystems'].rename(columns={'genes': genome_name})
        try:
            subsystems_df.drop('', level=0, axis=0, inplace=True)
            subsystems_df.drop('secondary-metabolites', level=0, axis=0, inplace=True)
        except Exception as e:
            pass

        if all_genome_feature_data is None:
            all_genome_feature_data = subsystems_df
        else:
            # all_genome_feature_data.to_excel(excel_file)
            # new_adds_excell = context.BASE_DIR / 'functional_genes_new.xls'
            #subsystems_df[genome_name].to_excel(new_adds_excell)
            all_genome_feature_data[genome_name] = subsystems_df[genome_name]
        all_genome_feature_data = all_genome_feature_data.copy()
    del all_genome_feature_data['profiles']
    all_genome_feature_data.to_excel(excel_file)


def main():
    print(f'This is metaerg.py {VERSION}')
    context.init(**parse_arguments().__dict__)

    if context.METAERG_MODE == context.METAERG_MODE_CREATE_DATABASE:
        context.log(f'Creating/installing/downloading metaerg databases. Tasks: {context.TASKS}; '
                    f'force: {context.FORCE}.')
        for db_installer in registry.DATABASE_INSTALLER_REGISTRY:
            db_installer()
    elif context.METAERG_MODE == context.METAERG_MODE_DOWNLOAD_DATABASE:
        context.log('Downloading premade databases from https://object-arbutus.cloud.computecanada.ca...')
        database_tarbal_file = context.DATABASE_DIR / 'metaerg_db_207_v2.tar.gz'
        context.download('https://object-arbutus.cloud.computecanada.ca/metaerg/metaerg_2.25_gtdb_207_v2.tar.gz',
                         database_tarbal_file)
        md5sum = md5(open(database_tarbal_file,'rb').read()).hexdigest()
        print(md5sum)
        if '48ffe7150711dd6f982e1f3d759e4ff9' == md5sum:
            context.log(f'checksum {md5sum} as expected.')
        else:
            raise Exception('Downloaded database has incorrect checksum - download failed. Aborting...')
        context.log('Now extracting databases from tar archive...')
        database_archive = tarfile.open(database_tarbal_file)
        database_archive.extractall(context.DATABASE_DIR)
        # database_tarbal_file.unlink()
        context.TASKS = 'SA'
        metaerg.run_and_read.diamond_and_blastn.compile_databases()
        metaerg.run_and_read.functional_genes.install_functional_gene_databases()
        metaerg.run_and_read.antismash.format_antismash_databases()
    elif context.METAERG_MODE == context.METAERG_MODE_INSTALL_DEPS:
        install_all_helper_programs(context.BIN_DIR, context.PATH_TO_SIGNALP, context.PATH_TO_TMHMM)
    else:
        functional_gene_configuration.init_functional_gene_config()
        annotation_summaries = {}
        if context.PARALLEL_ANNOTATIONS > 1:
            outcomes = {}
            with futures.ProcessPoolExecutor(max_workers=context.PARALLEL_ANNOTATIONS) as executor:
                for genome_name, contig_file in zip(context.GENOME_NAMES, context.CONTIG_FILES):
                    outcomes[genome_name] = executor.submit(annotate_genome, genome_name, contig_file)
            for genome_name, future in outcomes.items():
                try:
                    annotation_summaries[genome_name] = future.result()
                except Exception:
                    context.log(f'({genome_name}) Error while processing:')
                    raise
        else:
            for genome_name, contig_file in zip(context.GENOME_NAMES, context.CONTIG_FILES):
                annotation_summaries[genome_name] = annotate_genome(genome_name, contig_file)
        tmhmm.cleanup(context.BASE_DIR)
        context.log('Now writing all-genomes overview html...')
        html_all_genomes.write_html(context.HTML_DIR)
        context.log('Now writing excel files with functional genes per genome...')
        write_functional_genes_to_xls(annotation_summaries)
        context.log(f'Done. Thank you for using metaerg.py {VERSION}')


if __name__ == "__main__":
    main()
